# vim: set ft=python ts=4 sw=4 expandtab:
# ruff: noqa: ERA001
from unittest.mock import patch

import pytest
import responses
from arrow import Arrow
from requests import HTTPError
from responses import matchers
from responses.registries import OrderedRegistry
from tenacity import RetryError

from smartapp.interface import SignatureError, SmartAppDefinition, SmartAppDispatcherConfig, SmartAppRequestContext
from smartapp.signature import SignatureVerifier, retrieve_public_key

# These tests are built on sample data found in the Joyent specification.
# See: https://github.com/TritonDataCenter/node-http-signature/blob/master/http_signing.md#appendix-a---test-values
#
# Joyent provides a sample request and also a public key and some sample signatures based
# on this request.  The sample request is::
#
#    POST /foo?param=value&pet=dog HTTP/1.1
#    Host: example.com
#    Date: Thu, 05 Jan 2014 21:31:40 GMT
#    Content-Type: application/json
#    Digest: SHA-256=X48E9qOokqqrvdts8nOJRJN3OWDUoyWxBf7kbu9DBPE=
#    Content-Length: 18
#
#    {"hello": "world"}
#
# The public key is stored below in `PUBLIC_SIGNING_KEY`
#
# In Joyent's documentation, there are two cases: the "default" case and the "all
# headers" case.  Below, the expected results are DEFAULTS_* and ALL_HEADERS_*.
#
# As a santity check, you can also verify behavior using OpenSSL::
#
#    #!/bin/bash
#
#    # Create the signing text per definition in the spec
#    # Note newlines between items but not at the end
#    printf "(request-target): post /foo?param=value&pet=dog\n" > signing_string.txt
#    printf "host: example.com\n" >> signing_string.txt
#    printf "date: Thu, 05 Jan 2014 21:31:40 GMT\n" >> signing_string.txt
#    printf "content-type: application/json\n" >> signing_string.txt
#    printf "digest: SHA-256=X48E9qOokqqrvdts8nOJRJN3OWDUoyWxBf7kbu9DBPE=\n" >> signing_string.txt
#    printf "content-length: 18" >> signing_string.txt
#
#    # Create a SHA-256 signature in base64 format, the string from Joyent's example
#    printf "Ef7MlxLXoBovhil3AlyjtBwAL9g4TN3tibLj7uuNB3CROat/9KaeQ4hW2NiJ+pZ6HQEOx9vYZAyi+7cmIkmJszJCut5kQLAwuX+Ms/mUFvpKlSo9StS2bMXDBNjOh4Auj774GFj4gwjS+3NhFeoqyr/MuN6HsEnkvn6zdgfE2i0=" > signature.base64.joyent
#
#    # Create SHA-256 signature in base64 format, using OpenSSL and the signing string
#    cat signing_string.txt | openssl dgst -binary -sign private.pem -sha256 | openssl enc -base64 -A > signature.base64.openssl
#
#    # Check that Joyent example and OpenSSL give same result
#    diff -Naur signature.base64.joyent signature.base64.openssl
#
#    # Decode the bas64 format, leaving raw bytes
#    cat signature.base64.openssl | openssl enc -A -d -base64 > signature
#
#    # Use OpenSSL's verify signature mechanism to check the raw byte signature against the signing string
#    cat signing_string.txt | openssl dgst -sha256 -verify public.pem -signature ./signature
#
#    # Clean up temporary files
#    rm -f signing_string.txt signature.base64.* signature.*
#
# When you run this, you get should::
#
#     $ sh check.sh
#     Verified OK

METHOD = "POST"
HOST = "example.com"
PATH = "/foo?param=value&pet=dog"
REQUEST_TARGET = "post /foo?param=value&pet=dog"
DATE_STR = "Thu, 05 Jan 2014 21:31:40 GMT"
DATE_OBJ = Arrow(2014, 1, 5, 21, 31, 40, tzinfo="GMT")
BODY = '{"hello": "world"}'
CONTENT_TYPE = "application/json"
CONTENT_LENGTH = str(len(BODY))  # should be 18
DIGEST = "SHA-256=X48E9qOokqqrvdts8nOJRJN3OWDUoyWxBf7kbu9DBPE="

CORRELATION = "XXX"
KEY_ID = "Test"
ALGORITHM = "rsa-sha256"
CLOCK_SKEW = 300
KEYSERVER_URL = "https://key.smartthings.com"
KEY_DOWNLOAD_URL = f"{KEYSERVER_URL}/{KEY_ID}"
SMARTAPP_URL = f"https://{HOST}{PATH}"

# public key used for the sample Joyent signatures
PUBLIC_SIGNING_KEY = """
-----BEGIN PUBLIC KEY-----
MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDCFENGw33yGihy92pDjZQhl0C3
6rPJj+CvfSC8+q28hxA161QFNUd13wuCTUcq0Qd2qsBe/2hFyc2DCJJg0h1L78+6
Z4UMR7EOcpfdUE9Hf3m/hs+FUR45uBJeDK1HSFHD8bHKD6kv8FPGfJTotc+2xjJw
oYi+1hqp1fIekaxsyQIDAQAB
-----END PUBLIC KEY-----
""".strip()

# a valid public key, just not the one used in Joyent's sample data
WRONG_SIGNING_KEY = """
-----BEGIN PUBLIC KEY-----
MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDTt945+S+6dhkvVQXH4E4Dy6hc
KqQ5Z4FSvwfXnu23sZ15A9vx43imVJE5bS0H6n893nh9RNvrYp98nGcQLLVhvdTq
wIpl+cWCurqGtwkAAqmNjXwCbj69hUHGXtqX3Jn5MKB5IghjEDU4N3dFGoCWWycb
4l8BNIcle/5s7Vo9vwIDAQAB
-----END PUBLIC KEY-----
""".strip()

DEFAULT_AUTHORIZATION = """
Signature keyId="Test",algorithm="rsa-sha256",signature="jKyvPcxB4JbmYY4mByyBY7cZfNl4OW9HpFQlG7N4YcJPteKTu4MWCLyk+gIr0wDgqtLWf9NLpMAMimdfsH7FSWGfbMFSrsVTHNTk0rK3usrfFnti1dxsM4jl0kYJCKTGI/UWkqiaxwNiKqGcdlEDrTcUhhsFsOIo8VhddmZTZ8w="
""".strip()

DEFAULT_SIGNATURE = """
jKyvPcxB4JbmYY4mByyBY7cZfNl4OW9HpFQlG7N4YcJPteKTu4MWCLyk+gIr0wDgqtLWf9NLpMAMimdfsH7FSWGfbMFSrsVTHNTk0rK3usrfFnti1dxsM4jl0kYJCKTGI/UWkqiaxwNiKqGcdlEDrTcUhhsFsOIo8VhddmZTZ8w=
""".strip()

DEFAULT_SIGNING_STRING = """
date: Thu, 05 Jan 2014 21:31:40 GMT
""".strip()

DEFAULT_ORIGINAL_HEADERS = {
    "Host": HOST,
    "Date": DATE_STR,
    "Content-Type": CONTENT_TYPE,
    "Digest": DIGEST,
    "Content-Length": CONTENT_LENGTH,
    "Authorization": DEFAULT_AUTHORIZATION,
    "x-st-correlation": CORRELATION,
}

DEFAULT_SIGNING_HEADERS = "Date"

DEFAULT_SIGNING_ATTRIBUTES = {
    "keyId": KEY_ID,
    "headers": DEFAULT_SIGNING_HEADERS,
    "algorithm": ALGORITHM,
    "signature": DEFAULT_SIGNATURE,
}

ALL_HEADERS_AUTHORIZATION = """
Signature keyId="Test",algorithm="rsa-sha256",headers="(request-target) host date content-type digest content-length",signature="Ef7MlxLXoBovhil3AlyjtBwAL9g4TN3tibLj7uuNB3CROat/9KaeQ4hW2NiJ+pZ6HQEOx9vYZAyi+7cmIkmJszJCut5kQLAwuX+Ms/mUFvpKlSo9StS2bMXDBNjOh4Auj774GFj4gwjS+3NhFeoqyr/MuN6HsEnkvn6zdgfE2i0="
""".strip()

ALL_HEADERS_SIGNATURE = """
Ef7MlxLXoBovhil3AlyjtBwAL9g4TN3tibLj7uuNB3CROat/9KaeQ4hW2NiJ+pZ6HQEOx9vYZAyi+7cmIkmJszJCut5kQLAwuX+Ms/mUFvpKlSo9StS2bMXDBNjOh4Auj774GFj4gwjS+3NhFeoqyr/MuN6HsEnkvn6zdgfE2i0=
""".strip()

ALL_HEADERS_ORIGINAL_HEADERS = {
    "Host": HOST,
    "Date": DATE_STR,
    "Content-Type": CONTENT_TYPE,
    "Digest": DIGEST,
    "Content-Length": CONTENT_LENGTH,
    "Authorization": ALL_HEADERS_AUTHORIZATION,
}

ALL_HEADERS_SIGNING_HEADERS = "(request-target) host date content-type digest content-length"

ALL_HEADERS_SIGNING_ATTRIBUTES = {
    "keyId": KEY_ID,
    "headers": ALL_HEADERS_SIGNING_HEADERS,
    "algorithm": ALGORITHM,
    "signature": ALL_HEADERS_SIGNATURE,
}

ALL_HEADERS_SIGNING_STRING = """
(request-target): post /foo?param=value&pet=dog
host: example.com
date: Thu, 05 Jan 2014 21:31:40 GMT
content-type: application/json
digest: SHA-256=X48E9qOokqqrvdts8nOJRJN3OWDUoyWxBf7kbu9DBPE=
content-length: 18
""".strip()

TIMEOUT_MATCHER = matchers.request_kwargs_matcher({"timeout": 5.0})


def build_config(clock_skew_sec: int | None = CLOCK_SKEW) -> SmartAppDispatcherConfig:
    """Build configuration to test with."""
    return SmartAppDispatcherConfig(
        check_signatures=True,
        clock_skew_sec=clock_skew_sec,
        keyserver_url=KEYSERVER_URL,
    )


def build_definition(target_url: str = SMARTAPP_URL) -> SmartAppDefinition:
    """Build a SmartApp definition to test with."""
    return SmartAppDefinition(
        id="",
        name="",
        description="",
        target_url=target_url,
        permissions=[],
        config_pages=[],
    )


def build_context(headers: dict[str, str]) -> SmartAppRequestContext:
    """Build a request context to test with."""
    return SmartAppRequestContext(body=BODY, headers=headers)


class TestSignatureVerifier:
    def test_default_signature_attributes(self):
        # technically, if we can properly verify a valid or invalid a signature, these are irrelevant
        # however, checking them helps us confirm that we have the right inputs to the algorithm
        context = build_context(headers=DEFAULT_ORIGINAL_HEADERS)
        config = build_config()
        definition = build_definition()
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        assert verifier.correlation_id == CORRELATION
        assert verifier.body == BODY
        assert verifier.method == METHOD
        assert verifier.path == PATH
        assert verifier.request_target == REQUEST_TARGET
        assert verifier.date == DATE_OBJ
        assert verifier.authorization == DEFAULT_AUTHORIZATION
        assert verifier.signing_attributes == DEFAULT_SIGNING_ATTRIBUTES
        assert verifier.key_id == KEY_ID
        assert verifier.keyserver_url == KEYSERVER_URL
        assert verifier.algorithm == ALGORITHM
        assert verifier.signature == DEFAULT_SIGNATURE
        assert verifier.signing_string == DEFAULT_SIGNING_STRING

    def test_all_headers_signature_attributes(self):
        # technically, if we can properly verify a valid or invalid a signature, these are irrelevant
        # however, checking them helps us confirm that we have the right inputs to the algorithm
        context = build_context(headers=ALL_HEADERS_ORIGINAL_HEADERS)
        config = build_config()
        definition = build_definition()
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        assert verifier.correlation_id is None  # because the header isn't in ALL_HEADERS_ORIGINAL_HEADERS
        assert verifier.body == BODY
        assert verifier.method == METHOD
        assert verifier.path == PATH
        assert verifier.request_target == REQUEST_TARGET
        assert verifier.date == DATE_OBJ
        assert verifier.authorization == ALL_HEADERS_AUTHORIZATION
        assert verifier.signing_attributes == ALL_HEADERS_SIGNING_ATTRIBUTES
        assert verifier.key_id == KEY_ID
        assert verifier.keyserver_url == KEYSERVER_URL
        assert verifier.algorithm == ALGORITHM
        assert verifier.signature == ALL_HEADERS_SIGNATURE
        assert verifier.signing_string == ALL_HEADERS_SIGNING_STRING

    def test_header(self):
        headers = {
            "Date": DATE_STR,
            "Authorization": DEFAULT_AUTHORIZATION,
            "empty": "",
            "whitespace": "\t\n",
            "none": None,
        }
        context = build_context(headers=headers)
        config = build_config()
        definition = build_definition()
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        for header in ["DATE", "Date", "date"]:
            assert verifier.header(header) == DATE_STR
        for header in ["AUTHORIZATION", "Authorization", "authorization"]:
            assert verifier.header(header) == DEFAULT_AUTHORIZATION
        for header in ["missing", "empty", "whitespace", "none"]:
            with pytest.raises(SignatureError, match=f"Header not found: {header}"):
                verifier.header(header)

    @patch("smartapp.signature.arrow_now")
    @pytest.mark.parametrize(
        "date",
        [
            DATE_OBJ.shift(seconds=-CLOCK_SKEW),
            DATE_OBJ.shift(seconds=-CLOCK_SKEW).shift(seconds=1),
            DATE_OBJ.shift(seconds=-CLOCK_SKEW).shift(seconds=2),
            DATE_OBJ,
            DATE_OBJ.shift(seconds=CLOCK_SKEW).shift(seconds=-2),
            DATE_OBJ.shift(seconds=CLOCK_SKEW).shift(seconds=-1),
            DATE_OBJ.shift(seconds=CLOCK_SKEW),
        ],
    )
    def test_verify_date_valid(self, now, date):
        # with a clock skew set, anything inclusively within the configured skew is valid
        context = build_context(headers=DEFAULT_ORIGINAL_HEADERS)
        config = build_config()
        definition = build_definition()
        now.return_value = date
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        verifier.verify_date()

    @patch("smartapp.signature.arrow_now")
    @pytest.mark.parametrize(
        "date",
        [
            DATE_OBJ.shift(seconds=-CLOCK_SKEW).shift(seconds=-2),
            DATE_OBJ.shift(seconds=-CLOCK_SKEW).shift(seconds=-1),
            DATE_OBJ.shift(seconds=CLOCK_SKEW).shift(seconds=1),
            DATE_OBJ.shift(seconds=CLOCK_SKEW).shift(seconds=2),
        ],
    )
    def test_verify_date_invalid(self, now, date):
        # with a clock skew set, anything 1 second or more outside the configured skew is invalid
        context = build_context(headers=DEFAULT_ORIGINAL_HEADERS)
        config = build_config()
        definition = build_definition()
        now.return_value = date
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        with pytest.raises(SignatureError, match="Request date is not current") as e:
            verifier.verify_date()
        assert e.value.correlation_id == context.correlation_id

    @patch("smartapp.signature.arrow_now")
    @pytest.mark.parametrize(
        "date",
        [
            DATE_OBJ.shift(seconds=-CLOCK_SKEW).shift(seconds=-2),
            DATE_OBJ.shift(seconds=-CLOCK_SKEW).shift(seconds=-1),
            DATE_OBJ.shift(seconds=-CLOCK_SKEW),
            DATE_OBJ.shift(seconds=-CLOCK_SKEW).shift(seconds=1),
            DATE_OBJ.shift(seconds=-CLOCK_SKEW).shift(seconds=2),
            DATE_OBJ,
            DATE_OBJ.shift(seconds=CLOCK_SKEW).shift(seconds=-2),
            DATE_OBJ.shift(seconds=CLOCK_SKEW).shift(seconds=-1),
            DATE_OBJ.shift(seconds=CLOCK_SKEW),
            DATE_OBJ.shift(seconds=CLOCK_SKEW).shift(seconds=1),
            DATE_OBJ.shift(seconds=CLOCK_SKEW).shift(seconds=2),
        ],
    )
    def test_verify_date_no_skew(self, now, date):
        # with clock skew set to None (via no_skew_config) any date is valid
        context = build_context(headers=DEFAULT_ORIGINAL_HEADERS)
        config = build_config(clock_skew_sec=None)
        definition = build_definition()
        now.return_value = date
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        verifier.verify_date()

    @patch("smartapp.signature.retrieve_public_key")
    @patch("smartapp.signature.arrow_now")
    def test_default_verify_valid(self, now, retrieve):
        # with a valid key from the Joyent sample data, everything should line up and we should have a valid signature
        context = build_context(headers=DEFAULT_ORIGINAL_HEADERS)
        config = build_config()
        definition = build_definition()
        now.return_value = DATE_OBJ
        retrieve.return_value = PUBLIC_SIGNING_KEY
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        verifier.verify()
        retrieve.assert_called_once_with(KEYSERVER_URL, KEY_ID)

    @patch("smartapp.signature.retrieve_public_key")
    @patch("smartapp.signature.arrow_now")
    def test_all_headers_verify_valid(self, now, retrieve):
        # with a valid key from the Joyent sample data, everything should line up and we should have a valid signature
        context = build_context(headers=DEFAULT_ORIGINAL_HEADERS)
        config = build_config()
        definition = build_definition()
        now.return_value = DATE_OBJ
        retrieve.return_value = PUBLIC_SIGNING_KEY
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        verifier.verify()
        retrieve.assert_called_once_with(KEYSERVER_URL, KEY_ID)

    @patch("smartapp.signature.retrieve_public_key")
    @patch("smartapp.signature.arrow_now")
    def test_default_verify_key_download_failure(self, now, retrieve):
        # if we fail to download the key, the signature can never be valid
        context = build_context(headers=DEFAULT_ORIGINAL_HEADERS)
        config = build_config()
        definition = build_definition()
        now.return_value = DATE_OBJ
        retrieve.side_effect = HTTPError("failed to download")
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        with pytest.raises(SignatureError, match=rf"Failed to retrieve key \[{KEY_ID}\]") as e:
            verifier.verify()
        assert e.value.correlation_id == context.correlation_id

    @patch("smartapp.signature.retrieve_public_key")
    @patch("smartapp.signature.arrow_now")
    def test_all_headers_verify_download_failure(self, now, retrieve):
        # if we fail to download the key, the signature can never be valid
        context = build_context(headers=ALL_HEADERS_ORIGINAL_HEADERS)
        config = build_config()
        definition = build_definition()
        now.return_value = DATE_OBJ
        retrieve.side_effect = HTTPError("failed to download")
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        with pytest.raises(SignatureError, match=rf"Failed to retrieve key \[{KEY_ID}\]") as e:
            verifier.verify()
        assert e.value.correlation_id == context.correlation_id

    @patch("smartapp.signature.retrieve_public_key")
    @patch("smartapp.signature.arrow_now")
    def test_default_verify_invalid_key(self, now, retrieve):
        # with an invalid key, the signature can never be valid
        context = build_context(headers=DEFAULT_ORIGINAL_HEADERS)
        config = build_config()
        definition = build_definition()
        now.return_value = DATE_OBJ
        retrieve.return_value = "bogus"
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        with pytest.raises(SignatureError, match="Signature is not valid") as e:
            verifier.verify()
        assert e.value.correlation_id == context.correlation_id

    @patch("smartapp.signature.retrieve_public_key")
    @patch("smartapp.signature.arrow_now")
    def test_all_headers_verify_invalid_key(self, now, retrieve):
        # with an invalid key, the signature can never be valid
        context = build_context(headers=ALL_HEADERS_ORIGINAL_HEADERS)
        config = build_config()
        definition = build_definition()
        now.return_value = DATE_OBJ
        retrieve.return_value = "bogus"
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        with pytest.raises(SignatureError, match="Signature is not valid") as e:
            verifier.verify()
        assert e.value.correlation_id == context.correlation_id

    @patch("smartapp.signature.retrieve_public_key")
    @patch("smartapp.signature.arrow_now")
    def test_default_verify_wrong_key(self, now, retrieve):
        # with the wrong key, the signature can never be valid
        context = build_context(headers=DEFAULT_ORIGINAL_HEADERS)
        config = build_config()
        definition = build_definition()
        now.return_value = DATE_OBJ
        retrieve.return_value = WRONG_SIGNING_KEY
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        with pytest.raises(SignatureError, match="Signature is not valid") as e:
            verifier.verify()
        assert e.value.correlation_id == context.correlation_id

    @patch("smartapp.signature.retrieve_public_key")
    @patch("smartapp.signature.arrow_now")
    def test_all_headers_verify_wrong_key(self, now, retrieve):
        # with the wrong key, the signature can never be valid
        context = build_context(headers=ALL_HEADERS_ORIGINAL_HEADERS)
        config = build_config()
        definition = build_definition()
        now.return_value = DATE_OBJ
        retrieve.return_value = WRONG_SIGNING_KEY
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        with pytest.raises(SignatureError, match="Signature is not valid") as e:
            verifier.verify()
        assert e.value.correlation_id == context.correlation_id

    @patch("smartapp.signature.retrieve_public_key")
    @patch("smartapp.signature.arrow_now")
    def test_default_verify_mismatch(self, now, retrieve):
        # this has a Date header that does not match the Joyent sample, causing an invalid signature
        headers = DEFAULT_ORIGINAL_HEADERS.copy()
        headers["Date"] = "Tue, 14 Jun 2022 21:23:01 GMT"  # different date than in Joyent sample
        context = build_context(headers)
        config = build_config()
        definition = build_definition()
        now.return_value = DATE_OBJ
        retrieve.return_value = PUBLIC_SIGNING_KEY
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        assert verifier.signing_string != DEFAULT_SIGNING_STRING
        with pytest.raises(SignatureError, match="Request date is not current") as e:
            verifier.verify()
        assert e.value.correlation_id == context.correlation_id

    @patch("smartapp.signature.retrieve_public_key")
    @patch("smartapp.signature.arrow_now")
    def test_all_headers_verify_mismatch(self, now, retrieve):
        # the bad_definition contains a URL that doesn't match the Joyent sample
        # we will generate a bad signing string and hence the signature is invalid
        context = build_context(headers=ALL_HEADERS_ORIGINAL_HEADERS)
        config = build_config()
        definition = build_definition(target_url="https://whatever.com/smartapp")  # does not match Joyent's example
        now.return_value = DATE_OBJ
        retrieve.return_value = PUBLIC_SIGNING_KEY
        verifier = SignatureVerifier(context=context, config=config, definition=definition)
        assert verifier.signing_string != ALL_HEADERS_SIGNING_STRING
        with pytest.raises(SignatureError, match="Signature is not valid") as e:
            verifier.verify()
        assert e.value.correlation_id == context.correlation_id

    def test_authorization_bad(self):
        headers = DEFAULT_ORIGINAL_HEADERS.copy()
        headers["Authorization"] = "bogus"  # does not match expected format
        context = build_context(headers)
        config = build_config()
        definition = build_definition()
        with pytest.raises(SignatureError, match="Authorization header is not a signature") as e:
            SignatureVerifier(context=context, config=config, definition=definition)
        assert e.value.correlation_id == context.correlation_id

    @pytest.mark.parametrize("attribute", ["keyId", "algorithm", "signature"])
    def test_authorization_missing_attribute(self, attribute):
        headers = DEFAULT_ORIGINAL_HEADERS.copy()
        headers["Authorization"] = DEFAULT_AUTHORIZATION.replace(attribute, "bogus")
        context = build_context(headers)
        config = build_config()
        definition = build_definition()
        with pytest.raises(SignatureError, match=f"Signature does not contain: {attribute}") as e:
            SignatureVerifier(context=context, config=config, definition=definition)
        assert e.value.correlation_id == context.correlation_id

    def test_authorization_invalid_algorithm(self):
        headers = DEFAULT_ORIGINAL_HEADERS.copy()
        headers["Authorization"] = DEFAULT_AUTHORIZATION.replace("rsa-sha256", "bogus")
        context = build_context(headers)
        config = build_config()
        definition = build_definition()
        with pytest.raises(SignatureError, match="Algorithm not supported: bogus") as e:
            SignatureVerifier(context=context, config=config, definition=definition)
        assert e.value.correlation_id == context.correlation_id


class TestRetrievePublicKey:
    # This checks both the retry logic and the LRU cache.
    # Note that the LRU cache does not cache exceptions, only returned values.

    def test_retrieve_public_key_succeeds(self):
        # Note that this test does not use a leading slash for the key id
        with responses.RequestsMock() as r:
            r.get(
                url="https://whatever.com/key-succeeds",
                status=200,
                body="public-key",
                match=[TIMEOUT_MATCHER],
            )
            assert retrieve_public_key("https://whatever.com", "key-succeeds") == "public-key"
            assert retrieve_public_key("https://whatever.com", "key-succeeds") == "public-key"
            assert len(r.calls) == 1  # only one call is made, because the first response is cached

    def test_retrieve_public_key_fails(self):
        # Note that this test does not use a leading slash for the key id
        with responses.RequestsMock() as r:
            r.get(
                url="https://whatever.com/key-fails",
                status=500,
                match=[TIMEOUT_MATCHER],
            )
            with pytest.raises(RetryError):
                retrieve_public_key("https://whatever.com", "key-fails")
            assert len(r.calls) == 5  # five calls are made because failures are not cached

    def test_retrieve_public_key_retry(self):
        # Note that this test does use a leading slash for the key id
        with responses.RequestsMock(registry=OrderedRegistry) as r:
            r.get(
                url="https://whatever.com/key-retry",
                status=500,
                match=[TIMEOUT_MATCHER],
            )
            r.get(
                url="https://whatever.com/key-retry",
                status=200,
                body="public-key",
                match=[TIMEOUT_MATCHER],
            )
            assert retrieve_public_key("https://whatever.com", "/key-retry") == "public-key"
            assert retrieve_public_key("https://whatever.com", "/key-retry") == "public-key"
            assert len(r.calls) == 2  # two calls are made because of the initial retry; the second attempt is cached



# vim: set ft=python ts=4 sw=4 expandtab:

"""
Converter to serialize and deserialize lifecycle objects to various formats.
"""

import json
from enum import Enum
from typing import Any, TypeVar

import yaml
from arrow import Arrow
from arrow import get as arrow_get
from arrow import now as arrow_now
from attrs import fields, has
from cattrs import GenConverter
from cattrs.gen import make_dict_structure_fn, make_dict_unstructure_fn, override
from yaml import SafeDumper

from smartapp.interface import (
    CONFIG_SETTING_BY_TYPE,
    CONFIG_VALUE_BY_TYPE,
    REQUEST_BY_PHASE,
    ConfigSetting,
    ConfigSettingType,
    ConfigValue,
    ConfigValueType,
    LifecyclePhase,
    LifecycleRequest,
)

DATETIME_ZONE = "UTC"

DATETIME_SEC_EPOCH = "1970-01-01T00:00:00Z"  # date of the UNIX epoch, which sometimes seems to mean "no date"
DATETIME_SEC_LEN = len("YYYY-MM-DDTHH:MM:SSZ")  # like "2017-09-13T04:18:12Z"
DATETIME_SEC_FORMAT = "YYYY-MM-DD[T]HH:mm:ss[Z]"

DATETIME_MS_EPOCH = "1970-01-01T00:00:00.000Z"  # date of the UNIX epoch, which sometimes seems to mean "no date"
DATETIME_MS_LEN = len("YYYY-MM-DDTHH:MM:SS.SSSZ")  # like "2017-09-13T04:18:12.992Z"
DATETIME_MS_FORMAT = "YYYY-MM-DD[T]HH:mm:ss.SSS[Z]"

T = TypeVar("T")


# Configure SafeDumper to handle Enum values
yaml.add_multi_representer(Enum, lambda d, e: d.represent_str(e.value), Dumper=SafeDumper)


def serialize_datetime(datetime: Arrow) -> str:
    """Serialize an Arrow datetime to a string."""
    # Note that we always use the full millisecond timestamp here and always convert to UTC
    return datetime.to(DATETIME_ZONE).format(DATETIME_MS_FORMAT)


def deserialize_datetime(datetime: str) -> Arrow:
    """Deserialize a string into an Arrow datetime."""
    # Dates from SmartThings are not as reliable as I had hoped.  The samples show a
    # format including milliseconds.  Actual data (at least sometimes) comes without
    # milliseconds.  Further, some requests come with a UNIX epoch date (1970-01-01) which
    # I guess is probably what happens when no date was set by the device.  I'm choosing
    # to interpret that as "now".
    if datetime in {DATETIME_MS_EPOCH, DATETIME_SEC_EPOCH}:
        return arrow_now()
    if len(datetime) == DATETIME_MS_LEN:
        return arrow_get(datetime, DATETIME_MS_FORMAT, tzinfo=DATETIME_ZONE)
    if len(datetime) == DATETIME_SEC_LEN:
        return arrow_get(datetime, DATETIME_SEC_FORMAT, tzinfo=DATETIME_ZONE)
    raise ValueError(f"Unknown datetime format: {datetime}")


# noinspection PyMethodMayBeStatic
class StandardConverter(GenConverter):
    """
    Standard cattrs converter supporting both JSON and YAML and using camelCase for fields.
    """

    # Note: we need to inherit from GenConverter and not Converter because we use PEP563 (postponed) annotations
    # See: https://stackoverflow.com/a/72539298/2907667 and https://github.com/python-attrs/cattrs/issues/41

    # The factory hooks convert snake case to camel case
    # See: https://cattrs.readthedocs.io/en/latest/usage.html#using-factory-hooks

    def __init__(self) -> None:
        super().__init__()
        self.register_unstructure_hook_factory(has, self._unstructure_camel_case)
        self.register_structure_hook_factory(has, self._structure_camel_case)

    def to_json(self, obj: Any) -> str:
        """Serialize an object to JSON."""
        return json.dumps(self.unstructure(obj), indent="  ")

    def from_json(self, data: str, cls: type[T]) -> T:
        """Deserialize an object from JSON."""
        return self.structure(json.loads(data), cls)

    def to_yaml(self, obj: Any) -> str:
        """Serialize an object to YAML."""
        return yaml.safe_dump(self.unstructure(obj), sort_keys=False)

    def from_yaml(self, data: str, cls: type[T]) -> T:
        """Deserialize an object from YAML."""
        return self.structure(yaml.safe_load(data), cls)

    def _to_camel_case(self, name: str) -> str:  # noqa: PLR6301
        """Convert a snake_case attribute name to camelCase instead."""
        components = name.split("_")
        return components[0] + "".join(x.title() for x in components[1:])

    # I can't figure out any way to declare this so MyPy is happy, but it does function properly
    def _unstructure_camel_case(self, cls: type[T]):  # type: ignore[no-untyped-def] # noqa: ANN202
        """Automatic snake_case to camelCase conversion when serializing any class."""
        return make_dict_unstructure_fn(cls, self, **{a.name: override(rename=self._to_camel_case(a.name)) for a in fields(cls)})  # type: ignore[arg-type]

    # I can't figure out any way to declare this so MyPy is happy, but it does function properly
    def _structure_camel_case(self, cls: type[T]):  # type: ignore[no-untyped-def] # noqa: ANN202
        """Automatic snake_case to camelCase conversion when deserializing any class."""
        return make_dict_structure_fn(cls, self, **{a.name: override(rename=self._to_camel_case(a.name)) for a in fields(cls)})  # type: ignore[arg-type]


# noinspection PyMethodMayBeStatic
class SmartAppConverter(StandardConverter):
    """
    Cattrs converter to serialize/deserialize SmartApp-related classes, supporting both JSON and YAML.
    """

    def __init__(self) -> None:
        super().__init__()
        self.register_unstructure_hook(Arrow, self._unstructure_datetime)
        self.register_structure_hook(Arrow, self._structure_datetime)
        self.register_structure_hook(ConfigValue, self._structure_config_value)
        self.register_structure_hook(ConfigSetting, self._structure_config_setting)
        self.register_structure_hook(LifecycleRequest, self._structure_request)

    def _unstructure_datetime(self, datetime: Arrow) -> str:  # noqa: PLR6301
        """Serialize an Arrow datetime to a string."""
        return serialize_datetime(datetime)

    def _structure_datetime(self, datetime: str, _: type[Arrow]) -> Arrow:  # noqa: PLR6301
        """Deserialize a string into an Arrow datetime."""
        return deserialize_datetime(datetime)

    def _structure_config_value(self, data: dict[str, Any], _: type[ConfigValue]) -> ConfigValue:
        """Deserialize input data into a ConfigValue of the proper type."""
        try:
            value_type = ConfigValueType[data["valueType"]]
            return self.structure(data, CONFIG_VALUE_BY_TYPE[value_type])
        except KeyError as e:
            raise ValueError("Unknown config value type") from e

    def _structure_config_setting(self, data: dict[str, Any], _: type[ConfigSetting]) -> ConfigSetting:
        """Deserialize input data into a ConfigSetting of the proper type."""
        try:
            value_type = ConfigSettingType[data["type"]]
            return self.structure(data, CONFIG_SETTING_BY_TYPE[value_type])  # type: ignore[arg-type]
        except KeyError as e:
            raise ValueError("Unknown config setting type") from e

    def _structure_request(self, data: dict[str, Any], _: type[LifecycleRequest]) -> LifecycleRequest:
        """Deserialize input data into a LifecycleRequest of the proper type."""
        try:
            phase = LifecyclePhase[data["lifecycle"]]
            return self.structure(data, REQUEST_BY_PHASE[phase])  # type: ignore[arg-type]
        except KeyError as e:
            raise ValueError("Unknown lifecycle phase") from e


CONVERTER = SmartAppConverter()



# vim: set ft=python ts=4 sw=4 expandtab:

"""
Manage the requests and responses that are part of the SmartApp lifecycle.
"""

import logging
from json import JSONDecodeError

from attr import field, frozen

from smartapp.converter import CONVERTER
from smartapp.interface import (
    AbstractRequest,
    BadRequestError,
    ConfigPhase,
    ConfigurationInitResponse,
    ConfigurationPageResponse,
    ConfigurationRequest,
    ConfirmationRequest,
    ConfirmationResponse,
    EventRequest,
    EventResponse,
    InstallRequest,
    InstallResponse,
    InternalError,
    LifecycleRequest,
    LifecycleResponse,
    OauthCallbackRequest,
    OauthCallbackResponse,
    SmartAppConfigManager,
    SmartAppDefinition,
    SmartAppDispatcherConfig,
    SmartAppError,
    SmartAppEventHandler,
    SmartAppRequestContext,
    UninstallRequest,
    UninstallResponse,
    UpdateRequest,
    UpdateResponse,
)
from smartapp.signature import SignatureVerifier

_LOGGER = logging.getLogger(__name__)


@frozen(kw_only=True)
class StaticConfigManager(SmartAppConfigManager):
    """
    Configuration manager that operates on static data.

    This is the configuration manager used by default in the dispatcher.  It operates on
    a static set of config pages.  This sort of static definition is adequate for lots of
    SmartApps, but it doesn't work for some types of complex configuration, where the responses
    need to be generated dynamically.  In that case, you can implement your own configuration
    manager with that specialized behavior.
    """

    def handle_page(
        self, _request: ConfigurationRequest, definition: SmartAppDefinition, page_id: int
    ) -> ConfigurationPageResponse:
        """Handle a CONFIGURATION PAGE lifecycle request."""
        if not definition.config_pages:
            raise ValueError("Static configuration manager requires at least one configured page.")
        previous_page_id = None if page_id == 1 else page_id - 1
        next_page_id = None if page_id >= len(definition.config_pages) else page_id + 1
        complete = page_id >= len(definition.config_pages)
        try:
            page = definition.config_pages[page_id - 1]  # page_id is 1-based, but we need 0-based for array
            return self.build_page_response(
                name=page.page_name,
                page_id=page_id,
                previous_page_id=previous_page_id,
                next_page_id=next_page_id,
                complete=complete,
                sections=page.sections,
            )
        except IndexError as e:
            raise ValueError(f"Page not found: {page_id}") from e


@frozen(kw_only=True)
class SmartAppDispatcher:
    # noinspection PyUnresolvedReferences
    """
    Dispatcher to manage the requests and responses that are part of the SmartApp lifecycle.

    You must provide both a definition and an event handler, but in some cases the handler
    methods will probably be no-ops without any custom logic.  For more information, see
    `SmartAppEventHandler`.

    Attributes:
        definition(SmartAppDefinition): The static definition for the SmartApp
        event_handler(SmartAppEventHandler): Application event handler for SmartApp lifecycle events
    """

    definition: SmartAppDefinition
    event_handler: SmartAppEventHandler
    config: SmartAppDispatcherConfig = field(factory=SmartAppDispatcherConfig)
    manager: SmartAppConfigManager = field(factory=StaticConfigManager)

    def dispatch(self, context: SmartAppRequestContext) -> str:
        """
        Dispatch a request, responding to SmartThings and invoking callbacks as needed.

        Args:
            context(SmartAppRequestContext): The request context

        Returns:
            str: Response JSON payload that to be returned to the POST caller

        Raises:
            SmartAppError: If processing fails
        """
        try:
            if self.config.log_json:  # put this right at the top, so we've got an opportunity to debug unexpected data
                _LOGGER.debug("[%s] Raw JSON: \n%s", context.correlation_id, context.body)  # note: may contain secrets!
            request: LifecycleRequest = CONVERTER.from_json(context.body, LifecycleRequest)  # type: ignore[arg-type]
            _LOGGER.info("[%s] Handling %s request", context.correlation_id, request.lifecycle)
            _LOGGER.debug("[%s] Date: %s", context.correlation_id, context.date)
            _LOGGER.debug("[%s] Signature: %s", context.correlation_id, context.signature)  # note: signature not confidential
            _LOGGER.debug("[%s] Request: %s", context.correlation_id, request)  # note: secrets are not serialized in repr()
            if self.config.check_signatures:
                SignatureVerifier(context=context, config=self.config, definition=self.definition).verify()
            response = self._handle_request(context.correlation_id, request)
            return CONVERTER.to_json(response)
        except SmartAppError:
            raise
        except JSONDecodeError as e:
            raise BadRequestError("Invalid JSON", context.correlation_id) from e
        except ValueError as e:
            raise BadRequestError(f"{e}", context.correlation_id) from e
        except Exception as e:
            raise InternalError(f"{e}", context.correlation_id) from e

    def _handle_request(self, correlation_id: str | None, request: AbstractRequest) -> LifecycleResponse:  # noqa: PLR0911
        """Handle a lifecycle request, returning the appropriate response."""
        if isinstance(request, ConfirmationRequest):
            self.event_handler.handle_confirmation(correlation_id, request)
            return self._handle_confirmation_request(request)
        if isinstance(request, ConfigurationRequest):
            self.event_handler.handle_configuration(correlation_id, request)
            return self._handle_config_request(request)
        if isinstance(request, InstallRequest):
            self.event_handler.handle_install(correlation_id, request)
            return InstallResponse()
        if isinstance(request, UpdateRequest):
            self.event_handler.handle_update(correlation_id, request)
            return UpdateResponse()
        if isinstance(request, UninstallRequest):
            self.event_handler.handle_uninstall(correlation_id, request)
            return UninstallResponse()
        if isinstance(request, OauthCallbackRequest):
            self.event_handler.handle_oauth_callback(correlation_id, request)
            return OauthCallbackResponse()
        if isinstance(request, EventRequest):
            self.event_handler.handle_event(correlation_id, request)
            return EventResponse()
        raise ValueError("Unknown lifecycle event")

    def _handle_confirmation_request(self, request: ConfirmationRequest) -> ConfirmationResponse:
        """Handle a CONFIRMATION lifecycle request, logging data and returning an appropriate response."""
        _LOGGER.info("CONFIRMATION [%s]: [%s]", request.app_id, request.confirmation_data.confirmation_url)
        return ConfirmationResponse(target_url=self.definition.target_url)

    def _handle_config_request(self, request: ConfigurationRequest) -> ConfigurationInitResponse | ConfigurationPageResponse:
        """Handle a CONFIGURATION lifecycle request, returning an appropriate response."""
        if request.configuration_data.phase == ConfigPhase.INITIALIZE:
            return self.manager.handle_initialize(request, self.definition)
        # if request.configuration_data.phase == ConfigPhase.PAGE:
        page_id = int(request.configuration_data.page_id)
        return self.manager.handle_page(request, self.definition, page_id)



# vim: set ft=python ts=4 sw=4 expandtab:

from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from smartapp.converter import CONVERTER
from smartapp.dispatcher import SmartAppDispatcher
from smartapp.interface import (
    BadRequestError,
    ConfigInit,
    ConfigInitData,
    ConfigPage,
    ConfigPageData,
    ConfigPhase,
    ConfigRequestData,
    ConfigSection,
    ConfigurationInitResponse,
    ConfigurationPageResponse,
    ConfigurationRequest,
    ConfirmationRequest,
    ConfirmationResponse,
    DecimalSetting,
    EventRequest,
    EventResponse,
    InstallRequest,
    InstallResponse,
    InternalError,
    LifecyclePhase,
    OauthCallbackRequest,
    OauthCallbackResponse,
    ParagraphSetting,
    SignatureError,
    SmartAppConfigPage,
    SmartAppDefinition,
    SmartAppDispatcherConfig,
    SmartAppEventHandler,
    SmartAppRequestContext,
    UninstallRequest,
    UninstallResponse,
    UpdateRequest,
    UpdateResponse,
)
from tests.smartapp.testutil import load_dir

FIXTURE_DIR = Path(__file__).parent / "fixtures" / "samples"
REQUEST_DIR = FIXTURE_DIR / "request"

CLOCK_SKEW_SEC = 555
CORRELATION = "AAAA"
HEADERS = {"x-st-correlation": CORRELATION}


@pytest.fixture
def requests():
    return load_dir(REQUEST_DIR)


@pytest.fixture
def definition():
    return SmartAppDefinition(
        id="id",
        name="name",
        description="description",
        target_url="target_url",
        permissions=["permission"],
        config_pages=[
            SmartAppConfigPage(
                page_name="First page",
                sections=[
                    ConfigSection(
                        name="Section 1",
                        settings=[
                            ParagraphSetting(
                                id="paragraph-id",
                                name="paragraph-name",
                                description="paragraph-description",
                                default_value="paragraph-text",
                            ),
                        ],
                    )
                ],
            ),
            SmartAppConfigPage(
                page_name="Second page",
                sections=[
                    ConfigSection(
                        name="Section 2",
                        settings=[
                            DecimalSetting(
                                id="decimal-id",
                                name="decimal-name",
                                description="decimal-description",
                                required=False,
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )


@pytest.fixture
def event_handler() -> SmartAppEventHandler:
    return MagicMock()


@pytest.fixture
def dispatcher(definition: SmartAppDefinition, event_handler: SmartAppEventHandler) -> SmartAppDispatcher:
    return SmartAppDispatcher(
        definition=definition,
        event_handler=event_handler,
        config=SmartAppDispatcherConfig(check_signatures=False, clock_skew_sec=CLOCK_SKEW_SEC),
    )


@pytest.fixture
def dispatcher_with_check(definition: SmartAppDefinition, event_handler: SmartAppEventHandler) -> SmartAppDispatcher:
    return SmartAppDispatcher(
        definition=definition,
        event_handler=event_handler,
        config=SmartAppDispatcherConfig(check_signatures=True, clock_skew_sec=CLOCK_SKEW_SEC),
    )


# noinspection PyUnresolvedReferences
class TestSmartAppDispatcher:
    def test_invalid_json(self, dispatcher):
        # all of the requests have the same behavior, so we just check it once
        with pytest.raises(BadRequestError) as e:
            dispatcher.dispatch(SmartAppRequestContext(headers=HEADERS, body="bogus"))
        assert e.value.correlation_id == CORRELATION

    def test_wrong_json(self, dispatcher):
        # all of the requests have the same behavior, so we just check it once
        with pytest.raises(BadRequestError) as e:
            dispatcher.dispatch(SmartAppRequestContext(headers=HEADERS, body='{"hello":"world"}'))
        assert e.value.correlation_id == CORRELATION

    def test_handler_internal_error(self, requests, dispatcher):
        # all of the requests have the same behavior, so we just check it once
        request_json = requests["CONFIRMATION.json"]
        dispatcher.event_handler.handle_confirmation.side_effect = Exception("Hello")
        with pytest.raises(InternalError) as e:
            dispatcher.dispatch(SmartAppRequestContext(headers=HEADERS, body=request_json))
        assert e.value.correlation_id == CORRELATION

    @patch("smartapp.dispatcher.SignatureVerifier")
    def test_disabled_signature(self, signature_verifier, requests, dispatcher):
        # all of the requests have the same behavior, so we just check it once
        obj = MagicMock()
        obj.verify = MagicMock()
        signature_verifier.side_effect = obj
        request_json = requests["CONFIRMATION.json"]
        dispatcher.dispatch(SmartAppRequestContext(headers=HEADERS, body=request_json))
        signature_verifier.assert_not_called()
        obj.verify.assert_not_called()

    @patch("smartapp.dispatcher.SignatureVerifier")
    def test_enabled_signature(self, signature_verifier, requests, dispatcher_with_check):
        # all of the requests have the same behavior, so we just check it once
        obj = MagicMock()
        obj.verify = MagicMock()
        signature_verifier.return_value = obj
        request_json = requests["CONFIRMATION.json"]
        dispatcher_with_check.dispatch(SmartAppRequestContext(headers=HEADERS, body=request_json))
        signature_verifier.assert_called_once_with(
            context=SmartAppRequestContext(headers=HEADERS, body=request_json),
            config=dispatcher_with_check.config,
            definition=dispatcher_with_check.definition,
        )
        obj.verify.assert_called_once()

    @patch("smartapp.dispatcher.SignatureVerifier")
    def test_bad_signature(self, signature_verifier, requests, dispatcher_with_check):
        # all of the requests have the same behavior, so we just check it once
        exception = SignatureError("Hello")
        obj = MagicMock()
        obj.verify = MagicMock()
        obj.verify.side_effect = exception  # this what would be thrown, so make sure it comes through
        signature_verifier.return_value = obj
        request_json = requests["CONFIRMATION.json"]
        with pytest.raises(SignatureError) as e:
            dispatcher_with_check.dispatch(SmartAppRequestContext(headers=HEADERS, body=request_json))
        assert e.value is exception
        signature_verifier.assert_called_once_with(
            context=SmartAppRequestContext(headers=HEADERS, body=request_json),
            config=dispatcher_with_check.config,
            definition=dispatcher_with_check.definition,
        )
        obj.verify.assert_called_once()

    def test_confirmation(self, requests, dispatcher):
        request_json = requests["CONFIRMATION.json"]
        request = CONVERTER.from_json(request_json, ConfirmationRequest)
        response_json = dispatcher.dispatch(SmartAppRequestContext(headers=HEADERS, body=request_json))
        assert response_json == CONVERTER.to_json(ConfirmationResponse(target_url="target_url"))
        dispatcher.event_handler.handle_confirmation.assert_called_once_with(CORRELATION, request)

    def test_install(self, requests, dispatcher):
        request_json = requests["INSTALL.json"]
        request = CONVERTER.from_json(request_json, InstallRequest)
        response_json = dispatcher.dispatch(SmartAppRequestContext(headers=HEADERS, body=request_json))
        assert response_json == CONVERTER.to_json(InstallResponse())
        dispatcher.event_handler.handle_install.assert_called_once_with(CORRELATION, request)

    def test_update(self, requests, dispatcher):
        request_json = requests["UPDATE.json"]
        request = CONVERTER.from_json(request_json, UpdateRequest)
        response_json = dispatcher.dispatch(SmartAppRequestContext(headers=HEADERS, body=request_json))
        assert response_json == CONVERTER.to_json(UpdateResponse())
        dispatcher.event_handler.handle_update.assert_called_once_with(CORRELATION, request)

    def test_uninstall(self, requests, dispatcher):
        request_json = requests["UNINSTALL.json"]
        request = CONVERTER.from_json(request_json, UninstallRequest)
        response_json = dispatcher.dispatch(SmartAppRequestContext(headers=HEADERS, body=request_json))
        assert response_json == CONVERTER.to_json(UninstallResponse())
        dispatcher.event_handler.handle_uninstall.assert_called_once_with(CORRELATION, request)

    def test_oauth_callback(self, requests, dispatcher):
        request_json = requests["OAUTH_CALLBACK.json"]
        request = CONVERTER.from_json(request_json, OauthCallbackRequest)
        response_json = dispatcher.dispatch(SmartAppRequestContext(headers=HEADERS, body=request_json))
        assert response_json == CONVERTER.to_json(OauthCallbackResponse())
        dispatcher.event_handler.handle_oauth_callback.assert_called_once_with(CORRELATION, request)

    def test_event(self, requests, dispatcher):
        request_json = requests["EVENT-DEVICE.json"]
        request = CONVERTER.from_json(request_json, EventRequest)
        response_json = dispatcher.dispatch(SmartAppRequestContext(headers=HEADERS, body=request_json))
        assert response_json == CONVERTER.to_json(EventResponse())
        dispatcher.event_handler.handle_event.assert_called_once_with(CORRELATION, request)

    def test_configuration_initialize(self, dispatcher):
        request = ConfigurationRequest(
            lifecycle=LifecyclePhase.CONFIGURATION,
            execution_id="execution_id",
            locale="locale",
            version="version",
            configuration_data=ConfigRequestData(
                installed_app_id="installed_app_id",
                phase=ConfigPhase.INITIALIZE,
                page_id="",
                previous_page_id="",
                config={},
            ),
        )
        response = ConfigurationInitResponse(
            configuration_data=ConfigInitData(
                # this is all from the provided SmartApp definition
                initialize=ConfigInit(
                    id="id",
                    name="name",
                    description="description",
                    permissions=["permission"],
                    first_page_id="1",
                )
            )
        )
        request_json = CONVERTER.to_json(request)
        response_json = dispatcher.dispatch(SmartAppRequestContext(headers=HEADERS, body=request_json))
        assert response_json == CONVERTER.to_json(response)
        dispatcher.event_handler.handle_configuration.assert_called_once_with(CORRELATION, request)

    def test_configuration_page_1of2(self, dispatcher):
        request = ConfigurationRequest(
            lifecycle=LifecyclePhase.CONFIGURATION,
            execution_id="execution_id",
            locale="locale",
            version="version",
            configuration_data=ConfigRequestData(
                installed_app_id="installed_app_id",
                phase=ConfigPhase.PAGE,
                page_id="1",
                previous_page_id="",
                config={},  # TODO: not 100% clear what comes across in these page requests, examples are odd
            ),
        )
        response = ConfigurationPageResponse(
            configuration_data=ConfigPageData(
                # most of this comes from the provided SmartApp definition, but some is derived
                page=ConfigPage(
                    page_id="1",
                    name="First page",
                    previous_page_id=None,
                    next_page_id="2",
                    complete=False,
                    sections=[
                        ConfigSection(
                            name="Section 1",
                            settings=[
                                ParagraphSetting(
                                    id="paragraph-id",
                                    name="paragraph-name",
                                    description="paragraph-description",
                                    default_value="paragraph-text",
                                ),
                            ],
                        )
                    ],
                )
            )
        )
        request_json = CONVERTER.to_json(request)
        response_json = dispatcher.dispatch(SmartAppRequestContext(headers=HEADERS, body=request_json))
        assert response_json == CONVERTER.to_json(response)
        dispatcher.event_handler.handle_configuration.assert_called_once_with(CORRELATION, request)

    def test_configuration_page_2of2(self, dispatcher):
        request = ConfigurationRequest(
            lifecycle=LifecyclePhase.CONFIGURATION,
            execution_id="execution_id",
            locale="locale",
            version="version",
            configuration_data=ConfigRequestData(
                installed_app_id="installed_app_id",
                phase=ConfigPhase.PAGE,
                page_id="2",
                previous_page_id="",
                config={},  # TODO: not 100% clear what comes across in these page requests, examples are odd
            ),
        )
        response = ConfigurationPageResponse(
            configuration_data=ConfigPageData(
                # most of this comes from the provided SmartApp definition, but some is derived
                page=ConfigPage(
                    page_id="2",
                    name="Second page",
                    previous_page_id="1",
                    next_page_id=None,
                    complete=True,
                    sections=[
                        ConfigSection(
                            name="Section 2",
                            settings=[
                                DecimalSetting(
                                    id="decimal-id",
                                    name="decimal-name",
                                    description="decimal-description",
                                    required=False,
                                ),
                            ],
                        )
                    ],
                )
            )
        )
        request_json = CONVERTER.to_json(request)
        response_json = dispatcher.dispatch(SmartAppRequestContext(headers=HEADERS, body=request_json))
        assert response_json == CONVERTER.to_json(response)
        dispatcher.event_handler.handle_configuration.assert_called_once_with(CORRELATION, request)



# vim: set ft=python ts=4 sw=4 expandtab:

from json import JSONDecodeError
from pathlib import Path
from unittest.mock import patch

import pytest
from arrow import arrow

from smartapp.converter import CONVERTER, deserialize_datetime, serialize_datetime
from smartapp.interface import (
    BooleanSetting,
    BooleanValue,
    ConfigInit,
    ConfigInitData,
    ConfigPage,
    ConfigPageData,
    ConfigPhase,
    ConfigRequestData,
    ConfigSection,
    ConfigurationInitResponse,
    ConfigurationPageResponse,
    ConfigurationRequest,
    ConfirmationData,
    ConfirmationRequest,
    ConfirmationResponse,
    DecimalSetting,
    DeviceConfigValue,
    DeviceSetting,
    DeviceValue,
    EmailSetting,
    EnumOption,
    EnumOptionGroup,
    EnumSetting,
    Event,
    EventData,
    EventRequest,
    EventResponse,
    EventType,
    IconSetting,
    ImageSetting,
    InstallData,
    InstalledApp,
    InstallRequest,
    InstallResponse,
    LifecyclePhase,
    LifecycleRequest,
    LinkSetting,
    NumberSetting,
    OauthCallbackData,
    OauthCallbackRequest,
    OauthCallbackResponse,
    OauthSetting,
    PageSetting,
    ParagraphSetting,
    PhoneSetting,
    SmartAppConfigPage,
    SmartAppDefinition,
    SmartAppDispatcherConfig,
    StringConfigValue,
    StringValue,
    TextSetting,
    TimeSetting,
    UninstallData,
    UninstallRequest,
    UninstallResponse,
    UpdateData,
    UpdateRequest,
    UpdateResponse,
)
from tests.smartapp.testutil import load_dir

FIXTURE_DIR = Path(__file__).parent / "fixtures" / "samples"
REQUEST_DIR = FIXTURE_DIR / "request"
RESPONSE_DIR = FIXTURE_DIR / "response"
SETTINGS_DIR = FIXTURE_DIR / "settings"

NOW = arrow.Arrow(2022, 6, 1, 2, 3, 4, microsecond=5, tzinfo="UTC")


@pytest.fixture
def requests():
    return load_dir(REQUEST_DIR)


@pytest.fixture
def responses():
    return load_dir(RESPONSE_DIR)


@pytest.fixture
def settings():
    return load_dir(SETTINGS_DIR)


def validate_json_roundtrip(json, expected, cls):
    """Validate a JSON round trip."""
    if json:
        obj = CONVERTER.from_json(json, cls)
        assert obj == expected
    converted = CONVERTER.from_json(CONVERTER.to_json(expected), cls)
    assert converted == expected and converted is not expected


def validate_yaml_roundtrip(yaml, expected, cls):
    """Validate a YAML round trip."""
    if yaml:
        obj = CONVERTER.from_yaml(yaml, cls)
        assert obj == expected
    converted = CONVERTER.from_yaml(CONVERTER.to_yaml(expected), cls)
    assert converted == expected and converted is not expected


class TestDatetime:
    @pytest.mark.parametrize(
        "datetime,expected",
        [
            [arrow.Arrow(2017, 9, 13, 4, 18, 12, microsecond=469000, tzinfo="UTC"), "2017-09-13T04:18:12.469Z"],
            [arrow.Arrow(2017, 9, 13, 4, 18, 12, microsecond=469123, tzinfo="UTC"), "2017-09-13T04:18:12.469Z"],
            [arrow.Arrow(2017, 9, 13, 4, 18, 12, microsecond=0, tzinfo="UTC"), "2017-09-13T04:18:12.000Z"],
            [arrow.Arrow(1970, 1, 1, 0, 0, 0, microsecond=0, tzinfo="UTC"), "1970-01-01T00:00:00.000Z"],
        ],
    )
    def test_serialize_datetime(self, datetime, expected):
        assert serialize_datetime(datetime) == expected

    @patch("smartapp.converter.arrow_now")
    @pytest.mark.parametrize(
        "datetime,expected",
        [
            ["2017-09-13T04:18:12.469Z", arrow.Arrow(2017, 9, 13, 4, 18, 12, microsecond=469000, tzinfo="UTC")],
            ["2017-09-13T04:18:12.000Z", arrow.Arrow(2017, 9, 13, 4, 18, 12, microsecond=0, tzinfo="UTC")],
            ["2017-09-13T04:18:12Z", arrow.Arrow(2017, 9, 13, 4, 18, 12, microsecond=0, tzinfo="UTC")],
            ["2022-06-16T15:17:24.883Z", arrow.Arrow(2022, 6, 16, 10, 17, 24, microsecond=883000, tzinfo="America/Chicago")],
            ["2022-06-16T15:17:24.000Z", arrow.Arrow(2022, 6, 16, 10, 17, 24, microsecond=0, tzinfo="America/Chicago")],
            ["2022-06-16T15:16:24Z", arrow.Arrow(2022, 6, 16, 10, 16, 24, microsecond=0, tzinfo="America/Chicago")],
            ["1970-01-01T00:00:00.000Z", NOW],
            ["1970-01-01T00:00:00Z", NOW],
        ],
    )
    def test_deserialize_datetime(self, now, datetime, expected):
        now.return_value = NOW
        assert deserialize_datetime(datetime) == expected

    @pytest.mark.parametrize(
        "datetime",
        [
            "",  # empty
            "2017-09-13T04:18:12.469",  # no trailing Z
            "2017-09-13T04:18:12",  # no trailing Z
        ],
    )
    def test_deserialize_datetime_invalid(self, datetime):
        with pytest.raises(ValueError, match=f"Unknown datetime format: {datetime}"):
            deserialize_datetime(datetime)


class TestConvertSettings:
    def test_boolean(self, settings):
        json = settings["BOOLEAN.json"]
        expected = BooleanSetting(
            id="myBooleanSetting",
            name="True or false?",
            description="Tap to set",
            required=True,
            default_value=BooleanValue.TRUE,
        )
        validate_json_roundtrip(json, expected, BooleanSetting)
        validate_yaml_roundtrip(None, expected, BooleanSetting)

    def test_decimal(self, settings):
        json = settings["DECIMAL.json"]
        expected = DecimalSetting(
            id="myDecimalSetting",
            name="Enter a decimal value",
            description="Tap to set",
        )
        validate_json_roundtrip(json, expected, DecimalSetting)
        validate_yaml_roundtrip(None, expected, DecimalSetting)

    def test_device(self, settings):
        json = settings["DEVICE.json"]
        expected = DeviceSetting(
            id="contactSensor",
            name="Which contact sensor?",
            description="Tap to set",
            required=True,
            multiple=False,
            capabilities=["contactSensor"],
            permissions=["r"],
        )
        validate_json_roundtrip(json, expected, DeviceSetting)
        validate_yaml_roundtrip(None, expected, DeviceSetting)

    def test_email(self, settings):
        json = settings["EMAIL.json"]
        expected = EmailSetting(
            id="myEmailSetting",
            name="Enter an email address",
            description="Tap to set",
        )
        validate_json_roundtrip(json, expected, EmailSetting)
        validate_yaml_roundtrip(None, expected, EmailSetting)

    def test_enum(self, settings):
        json = settings["ENUM.json"]
        expected = EnumSetting(
            id="myEnumSetting",
            name="Choose what applies",
            description="Tap to set",
            required=True,
            multiple=True,
            options=[EnumOption(id="option-1", name="Option 1"), EnumOption(id="option-2", name="Option 2")],
        )
        validate_json_roundtrip(json, expected, EnumSetting)
        validate_yaml_roundtrip(None, expected, EnumSetting)

    def test_enum_group(self, settings):
        json = settings["ENUM-GROUP.json"]
        expected = EnumSetting(
            id="myGroupedEnumSetting",
            name="Choose what applies",
            description="Tap to set",
            required=True,
            multiple=True,
            grouped_options=[
                EnumOptionGroup(
                    name="Group 1", options=[EnumOption(id="option-1", name="Option 1"), EnumOption(id="option-2", name="Option 2")]
                ),
                EnumOptionGroup(
                    name="Group 2",
                    options=[
                        EnumOption(id="option-3", name="Option 3"),
                        EnumOption(id="option-4", name="Option 4"),
                        EnumOption(id="option-5", name="Option 5"),
                    ],
                ),
            ],
        )
        validate_json_roundtrip(json, expected, EnumSetting)
        validate_yaml_roundtrip(None, expected, EnumSetting)

    def test_icon(self, settings):
        json = settings["ICON.json"]
        expected = IconSetting(
            id="myIconInput",
            name="Some icon information",
            description="Some description",
            image="https://some-image-url",
        )
        validate_json_roundtrip(json, expected, IconSetting)
        validate_yaml_roundtrip(None, expected, IconSetting)

    def test_image(self, settings):
        json = settings["IMAGE.json"]
        expected = ImageSetting(
            id="myImageInput",
            name="Choose what applies",
            description="Tap to set",
            image="https://some-image-url",
        )
        validate_json_roundtrip(json, expected, ImageSetting)
        validate_yaml_roundtrip(None, expected, ImageSetting)

    def test_link(self, settings):
        json = settings["LINK.json"]
        expected = LinkSetting(
            id="myLinkSetting",
            name="Visit the following link",
            description="Tap to visit",
            url="https://some-site-url",
            image="https://some-image-url",
        )
        validate_json_roundtrip(json, expected, LinkSetting)
        validate_yaml_roundtrip(None, expected, LinkSetting)

    def test_number(self, settings):
        json = settings["NUMBER.json"]
        expected = NumberSetting(
            id="myNumberSetting",
            name="Enter a number",
            description="Tap to set",
        )
        validate_json_roundtrip(json, expected, NumberSetting)
        validate_yaml_roundtrip(None, expected, NumberSetting)

    def test_oauth(self, settings):
        json = settings["OAUTH.json"]
        expected = OauthSetting(
            id="myOauthSetting",
            name="Authenticate with the third party service",
            description="Tap to set",
            browser=False,
            url_template="http://www.some-third-party.com/oauth?param1=1&param2=2&callback=https%3A%2F%2Fapi.smartthings.com%2Foauth%2Fcallback",
        )
        validate_json_roundtrip(json, expected, OauthSetting)
        validate_yaml_roundtrip(None, expected, OauthSetting)

    def test_page(self, settings):
        json = settings["PAGE.json"]
        expected = PageSetting(
            id="myPageSetting",
            name="Choose what applies",
            description="Tap to set",
            page="page-id",
            image="https://some-image-url",
        )
        validate_json_roundtrip(json, expected, PageSetting)
        validate_yaml_roundtrip(None, expected, PageSetting)

    def test_paragraph(self, settings):
        json = settings["PARAGRAPH.json"]
        expected = ParagraphSetting(
            id="myParagraphSetting",
            name="Some information title",
            description="Some description",
            default_value="This is the information to display.",
        )
        validate_json_roundtrip(json, expected, ParagraphSetting)
        validate_yaml_roundtrip(None, expected, ParagraphSetting)

    def test_phone(self, settings):
        json = settings["PHONE.json"]
        expected = PhoneSetting(
            id="myPhoneSetting",
            name="Enter a phone number",
            description="Tap to set",
        )
        validate_json_roundtrip(json, expected, PhoneSetting)
        validate_yaml_roundtrip(None, expected, PhoneSetting)

    def test_text(self, settings):
        json = settings["TEXT.json"]
        expected = TextSetting(
            id="myTextSetting",
            name="Enter some text",
            description="Tap to set",
            required=True,
            default_value="Some default value",
        )
        validate_json_roundtrip(json, expected, TextSetting)
        validate_yaml_roundtrip(None, expected, TextSetting)

    def test_time(self, settings):
        json = settings["TIME.json"]
        expected = TimeSetting(
            id="myTimeInput",
            name="Choose a time",
            description="Tap to set",
        )
        validate_json_roundtrip(json, expected, TimeSetting)
        validate_yaml_roundtrip(None, expected, TimeSetting)


class TestStringSecrets:
    # Some requests contain secret tokens in the auth_token or refresh_token fields
    # These must not be included when turning the object into a string, which attrs handles with field(repr=False)

    @pytest.mark.parametrize(
        "source",
        [
            "INSTALL.json",
            "UPDATE.json",
            "EVENT-TIMER.json",
            "EVENT-DEVICE.json",
        ],
    )
    def test_secrets(self, requests, source):
        json = requests[source]
        request = CONVERTER.from_json(json, LifecycleRequest)
        for string in ["%s" % request, f"{request}", str(request), repr(request)]:  # noqa: UP031
            assert "auth_token" not in string and "authTokenValue" not in string
            assert "refresh_token" not in string and "refreshTokenValue" not in string


class TestConvertResponses:
    def test_confirmation(self, responses):
        json = responses["CONFIRMATION.json"]
        expected = ConfirmationResponse(target_url="{TARGET_URL}")
        validate_json_roundtrip(json, expected, ConfirmationResponse)
        validate_yaml_roundtrip(None, expected, ConfirmationResponse)

    def test_configuration_initialize(self, responses):
        # Response for INITIALIZE is different than for PAGE
        json = responses["CONFIGURATION-INITIALIZE.json"]
        expected = ConfigurationInitResponse(
            configuration_data=ConfigInitData(
                initialize=ConfigInit(
                    name="On When Open/Off When Shut WebHook App",
                    description="On When Open/Off When Shut WebHook App",
                    id="app",
                    permissions=[],
                    first_page_id="1",
                )
            )
        )
        validate_json_roundtrip(json, expected, ConfigurationInitResponse)
        validate_yaml_roundtrip(None, expected, ConfigurationInitResponse)

    def test_configuration_page_only(self, responses):
        # Response for PAGE is different than for INITIALIZE
        # This tests the example for an only page (1 of 1)
        json = responses["CONFIGURATION-PAGE-only.json"]
        expected = ConfigurationPageResponse(
            configuration_data=ConfigPageData(
                page=ConfigPage(
                    page_id="1",
                    name="On When Open/Off When Shut WebHook App",
                    next_page_id=None,
                    previous_page_id=None,
                    complete=True,
                    sections=[
                        ConfigSection(
                            name="When this opens/closes...",
                            settings=[
                                DeviceSetting(
                                    id="contactSensor",
                                    name="Which contact sensor?",
                                    description="Tap to set",
                                    required=True,
                                    multiple=False,
                                    capabilities=["contactSensor"],
                                    permissions=["r"],
                                )
                            ],
                        ),
                        ConfigSection(
                            name="Turn on/off this light...",
                            settings=[
                                DeviceSetting(
                                    id="lightSwitch",
                                    name="Which switch?",
                                    description="Tap to set",
                                    required=True,
                                    multiple=False,
                                    capabilities=["switch"],
                                    permissions=["r", "x"],
                                )
                            ],
                        ),
                    ],
                )
            )
        )
        validate_json_roundtrip(json, expected, ConfigurationPageResponse)
        validate_yaml_roundtrip(None, expected, ConfigurationPageResponse)

    def test_configuration_page_1of2(self, responses):
        # Response for PAGE is different than for INITIALIZE
        # This tests the example for page 1 of 2
        json = responses["CONFIGURATION-PAGE-1of2.json"]
        expected = ConfigurationPageResponse(
            configuration_data=ConfigPageData(
                page=ConfigPage(
                    page_id="1",
                    name="On When Open/Off When Shut WebHook App",
                    next_page_id="2",
                    previous_page_id=None,
                    complete=False,
                    sections=[
                        ConfigSection(
                            name="When this opens/closes...",
                            settings=[
                                DeviceSetting(
                                    id="contactSensor",
                                    name="Which contact sensor?",
                                    description="Tap to set",
                                    required=True,
                                    multiple=False,
                                    capabilities=["contactSensor"],
                                    permissions=["r"],
                                )
                            ],
                        ),
                    ],
                )
            )
        )
        validate_json_roundtrip(json, expected, ConfigurationPageResponse)
        validate_yaml_roundtrip(None, expected, ConfigurationPageResponse)

    def test_install(self, responses):
        json = responses["INSTALL.json"]
        expected = InstallResponse()
        validate_json_roundtrip(json, expected, InstallResponse)
        validate_yaml_roundtrip(None, expected, InstallResponse)

    def test_update(self, responses):
        json = responses["UPDATE.json"]
        expected = UpdateResponse()
        validate_json_roundtrip(json, expected, UpdateResponse)
        validate_yaml_roundtrip(None, expected, UpdateResponse)

    def test_uninstall(self, responses):
        json = responses["UNINSTALL.json"]
        expected = UninstallResponse()
        validate_json_roundtrip(json, expected, UninstallResponse)
        validate_yaml_roundtrip(None, expected, UninstallResponse)

    def test_oauth_callback(self, responses):
        json = responses["OAUTH_CALLBACK.json"]
        expected = OauthCallbackResponse()
        validate_json_roundtrip(json, expected, OauthCallbackResponse)
        validate_yaml_roundtrip(None, expected, OauthCallbackResponse)

    def test_event(self, responses):
        json = responses["EVENT.json"]
        expected = EventResponse()
        validate_json_roundtrip(json, expected, EventResponse)
        validate_yaml_roundtrip(None, expected, EventResponse)


class TestConvertRequests:
    def test_invalid_json(self):
        # This is not a valid JSON string
        with pytest.raises(JSONDecodeError):
            CONVERTER.from_json("*(&)(&_)()", LifecycleRequest)

    def test_mismatch_json(self):
        # This is valid JSON, just not valid LifecycleRequest
        with pytest.raises(ValueError):
            CONVERTER.from_json('{"hello": "there"}', LifecycleRequest)

    def test_confirmation(self, requests):
        json = requests["CONFIRMATION.json"]
        expected = ConfirmationRequest(
            lifecycle=LifecyclePhase.CONFIRMATION,
            execution_id="8F8FA33E-2A5B-4BC5-826C-4B2AB73FE9DD",
            app_id="fd9949ee-a3bf-4069-b4b3-3e9c1c922e29",
            locale="en",
            version="0.1.0",
            confirmation_data=ConfirmationData(
                app_id="fd9949ee-a3bf-4069-b4b3-3e9c1c922e29", confirmation_url="{CONFIRMATION_URL}"
            ),
            settings={
                "property1": "string",
                "property2": "string",
            },
        )
        validate_json_roundtrip(json, expected, LifecycleRequest)
        validate_yaml_roundtrip(None, expected, LifecycleRequest)

    def test_configuration_initialize(self, requests):
        json = requests["CONFIGURATION-INITIALIZE.json"]
        expected = ConfigurationRequest(
            lifecycle=LifecyclePhase.CONFIGURATION,
            execution_id="b328f242-c602-4204-8d73-33c48ae180af",
            locale="en",
            version="1.0.0",
            configuration_data=ConfigRequestData(
                installed_app_id="string",
                phase=ConfigPhase.INITIALIZE,
                page_id="string",
                previous_page_id="string",
                config={
                    "property1": [
                        DeviceConfigValue(
                            device_config=DeviceValue(
                                device_id="31192dc9-eb45-4d90-b606-21e9b66d8c2b",
                                component_id="main",
                            ),
                        )
                    ],
                    "property2": [
                        DeviceConfigValue(
                            device_config=DeviceValue(
                                device_id="31192dc9-eb45-4d90-b606-21e9b66d8c2b",
                                component_id="main",
                            ),
                        )
                    ],
                },
            ),
            settings={
                "property1": "string",
                "property2": "string",
            },
        )
        validate_json_roundtrip(json, expected, LifecycleRequest)
        validate_yaml_roundtrip(None, expected, LifecycleRequest)

    def test_configuration_page(self, requests):
        # Note that the invalid "app" config item is ignored when we deserialize
        json = requests["CONFIGURATION-PAGE.json"]
        expected = ConfigurationRequest(
            lifecycle=LifecyclePhase.CONFIGURATION,
            execution_id="ce3975c1-0d03-3777-5250-0e61b15ad1d4",
            locale="en",
            version="0.1.0",
            configuration_data=ConfigRequestData(
                installed_app_id="58065067-52b9-49a5-a378-ce3871bc710b",
                phase=ConfigPhase.PAGE,
                page_id="2",
                previous_page_id="1",
                config={
                    "minutes": [
                        StringConfigValue(
                            string_config=StringValue(value="3"),
                        )
                    ],
                },
            ),
            settings={},
        )
        validate_json_roundtrip(json, expected, LifecycleRequest)
        validate_yaml_roundtrip(None, expected, LifecycleRequest)

    def test_install(self, requests):
        json = requests["INSTALL.json"]
        expected = InstallRequest(
            lifecycle=LifecyclePhase.INSTALL,
            execution_id="b328f242-c602-4204-8d73-33c48ae180af",
            locale="en",
            version="1.0.0",
            install_data=InstallData(
                auth_token="authTokenValue",  # noqa: S106 # this is not a real secret
                refresh_token="refreshTokenValue",  # noqa: S106 # this is not a real secret
                installed_app=InstalledApp(
                    installed_app_id="d692699d-e7a6-400d-a0b7-d5be96e7a564",
                    location_id="e675a3d9-2499-406c-86dc-8a492a886494",
                    config={
                        "contactSensor": [
                            DeviceConfigValue(
                                device_config=DeviceValue(
                                    device_id="e457978e-5e37-43e6-979d-18112e12c961",
                                    component_id="main",
                                ),
                            )
                        ],
                        "lightSwitch": [
                            DeviceConfigValue(
                                device_config=DeviceValue(
                                    device_id="74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                                    component_id="main",
                                ),
                            )
                        ],
                        "minutes": [
                            StringConfigValue(
                                string_config=StringValue(value="5"),
                            )
                        ],
                    },
                    permissions=[
                        "r:devices:e457978e-5e37-43e6-979d-18112e12c961",
                        "r:devices:74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                        "x:devices:74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                    ],
                ),
            ),
            settings={
                "property1": "string",
                "property2": "string",
            },
        )
        validate_json_roundtrip(json, expected, LifecycleRequest)
        validate_yaml_roundtrip(None, expected, LifecycleRequest)
        assert expected.token() == "authTokenValue"
        assert expected.app_id() == "d692699d-e7a6-400d-a0b7-d5be96e7a564"
        assert expected.location_id() == "e675a3d9-2499-406c-86dc-8a492a886494"

    def test_update(self, requests):
        json = requests["UPDATE.json"]
        expected = UpdateRequest(
            lifecycle=LifecyclePhase.UPDATE,
            execution_id="b328f242-c602-4204-8d73-33c48ae180af",
            locale="en",
            version="1.0.0",
            update_data=UpdateData(
                auth_token="authTokenValue",  # noqa: S106 # this is not a real secret
                refresh_token="refreshTokenValue",  # noqa: S106 # this is not a real secret
                installed_app=InstalledApp(
                    installed_app_id="d692699d-e7a6-400d-a0b7-d5be96e7a564",
                    location_id="e675a3d9-2499-406c-86dc-8a492a886494",
                    config={
                        "contactSensor": [
                            DeviceConfigValue(
                                device_config=DeviceValue(
                                    device_id="e457978e-5e37-43e6-979d-18112e12c961",
                                    component_id="main",
                                ),
                            )
                        ],
                        "lightSwitch": [
                            DeviceConfigValue(
                                device_config=DeviceValue(
                                    device_id="74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                                    component_id="main",
                                ),
                            )
                        ],
                        "minutes": [
                            StringConfigValue(
                                string_config=StringValue(value="5"),
                            )
                        ],
                    },
                    permissions=[
                        "r:devices:e457978e-5e37-43e6-979d-18112e12c961",
                        "r:devices:74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                        "x:devices:74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                    ],
                ),
                previous_config={
                    "contactSensor": [
                        DeviceConfigValue(
                            device_config=DeviceValue(
                                device_id="e457978e-5e37-43e6-979d-18112e12c961",
                                component_id="main",
                            ),
                        )
                    ],
                    "lightSwitch": [
                        DeviceConfigValue(
                            device_config=DeviceValue(
                                device_id="74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                                component_id="main",
                            ),
                        )
                    ],
                    "minutes": [
                        StringConfigValue(
                            string_config=StringValue(value="5"),
                        )
                    ],
                },
                previous_permissions=[
                    "r:devices:e457978e-5e37-43e6-979d-18112e12c961",
                    "r:devices:74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                    "x:devices:74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                ],
            ),
            settings={
                "property1": "string",
                "property2": "string",
            },
        )
        validate_json_roundtrip(json, expected, LifecycleRequest)
        validate_yaml_roundtrip(None, expected, LifecycleRequest)
        assert expected.token() == "authTokenValue"
        assert expected.app_id() == "d692699d-e7a6-400d-a0b7-d5be96e7a564"
        assert expected.location_id() == "e675a3d9-2499-406c-86dc-8a492a886494"

    def test_uninstall(self, requests):
        json = requests["UNINSTALL.json"]
        expected = UninstallRequest(
            lifecycle=LifecyclePhase.UNINSTALL,
            execution_id="b328f242-c602-4204-8d73-33c48ae180af",
            locale="en",
            version="1.0.0",
            uninstall_data=UninstallData(
                installed_app=InstalledApp(
                    installed_app_id="d692699d-e7a6-400d-a0b7-d5be96e7a564",
                    location_id="e675a3d9-2499-406c-86dc-8a492a886494",
                    config={
                        "contactSensor": [
                            DeviceConfigValue(
                                device_config=DeviceValue(
                                    device_id="e457978e-5e37-43e6-979d-18112e12c961",
                                    component_id="main",
                                ),
                            )
                        ],
                        "lightSwitch": [
                            DeviceConfigValue(
                                device_config=DeviceValue(
                                    device_id="74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                                    component_id="main",
                                ),
                            )
                        ],
                        "minutes": [
                            StringConfigValue(
                                string_config=StringValue(value="5"),
                            )
                        ],
                    },
                    permissions=[
                        "r:devices:e457978e-5e37-43e6-979d-18112e12c961",
                        "r:devices:74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                        "x:devices:74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                    ],
                )
            ),
            settings={
                "property1": "string",
                "property2": "string",
            },
        )
        validate_json_roundtrip(json, expected, LifecycleRequest)
        validate_yaml_roundtrip(None, expected, LifecycleRequest)
        assert expected.app_id() == "d692699d-e7a6-400d-a0b7-d5be96e7a564"
        assert expected.location_id() == "e675a3d9-2499-406c-86dc-8a492a886494"

    def test_oauth_callback(self, requests):
        json = requests["OAUTH_CALLBACK.json"]
        expected = OauthCallbackRequest(
            lifecycle=LifecyclePhase.OAUTH_CALLBACK,
            execution_id="b328f242-c602-4204-8d73-33c48ae180af",
            locale="en",
            version="1.0.0",
            o_auth_callback_data=OauthCallbackData(installed_app_id="string", url_path="string"),
        )
        validate_json_roundtrip(json, expected, LifecycleRequest)
        validate_yaml_roundtrip(None, expected, LifecycleRequest)

    def test_event_device(self, requests):
        json = requests["EVENT-DEVICE.json"]
        expected = EventRequest(
            lifecycle=LifecyclePhase.EVENT,
            execution_id="b328f242-c602-4204-8d73-33c48ae180af",
            locale="en",
            version="1.0.0",
            event_data=EventData(
                auth_token="authTokenValue",  # noqa: S106 # this is not a real secret
                installed_app=InstalledApp(
                    installed_app_id="d692699d-e7a6-400d-a0b7-d5be96e7a564",
                    location_id="e675a3d9-2499-406c-86dc-8a492a886494",
                    config={
                        "contactSensor": [
                            DeviceConfigValue(
                                device_config=DeviceValue(
                                    device_id="e457978e-5e37-43e6-979d-18112e12c961",
                                    component_id="main",
                                ),
                            )
                        ],
                        "lightSwitch": [
                            DeviceConfigValue(
                                device_config=DeviceValue(
                                    device_id="74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                                    component_id="main",
                                ),
                            )
                        ],
                        "minutes": [
                            StringConfigValue(
                                string_config=StringValue(value="5"),
                            )
                        ],
                    },
                    permissions=[
                        "r:devices:e457978e-5e37-43e6-979d-18112e12c961",
                        "r:devices:74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                        "x:devices:74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                    ],
                ),
                events=[
                    Event(
                        event_type=EventType.DEVICE_EVENT,
                        device_event={
                            "subscriptionName": "motion_sensors",
                            "eventId": "736e3903-001c-4d40-b408-ff40d162a06b",
                            "locationId": "499e28ba-b33b-49c9-a5a1-cce40e41f8a6",
                            "deviceId": "6f5ea629-4c05-4a90-a244-cc129b0a80c3",
                            "componentId": "main",
                            "capability": "motionSensor",
                            "attribute": "motion",
                            "value": "active",
                            "stateChange": True,
                        },
                    )
                ],
            ),
            settings={
                "property1": "string",
                "property2": "string",
            },
        )
        validate_json_roundtrip(json, expected, LifecycleRequest)
        validate_yaml_roundtrip(None, expected, LifecycleRequest)
        assert expected.token() == "authTokenValue"
        assert expected.app_id() == "d692699d-e7a6-400d-a0b7-d5be96e7a564"
        assert expected.location_id() == "e675a3d9-2499-406c-86dc-8a492a886494"

    def test_event_timer(self, requests):
        json = requests["EVENT-TIMER.json"]
        expected = EventRequest(
            lifecycle=LifecyclePhase.EVENT,
            execution_id="b328f242-c602-4204-8d73-33c48ae180af",
            locale="en",
            version="1.0.0",
            event_data=EventData(
                auth_token="authTokenValue",  # noqa: S106 # this is not a real secret
                installed_app=InstalledApp(
                    installed_app_id="d692699d-e7a6-400d-a0b7-d5be96e7a564",
                    location_id="e675a3d9-2499-406c-86dc-8a492a886494",
                    config={
                        "contactSensor": [
                            DeviceConfigValue(
                                device_config=DeviceValue(
                                    device_id="e457978e-5e37-43e6-979d-18112e12c961",
                                    component_id="main",
                                ),
                            )
                        ],
                        "lightSwitch": [
                            DeviceConfigValue(
                                device_config=DeviceValue(
                                    device_id="74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                                    component_id="main",
                                ),
                            )
                        ],
                        "minutes": [
                            StringConfigValue(
                                string_config=StringValue(value="5"),
                            )
                        ],
                    },
                    permissions=[
                        "r:devices:e457978e-5e37-43e6-979d-18112e12c961",
                        "r:devices:74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                        "x:devices:74aac3bb-91f2-4a88-8c49-ae5e0a234d76",
                    ],
                ),
                events=[
                    Event(
                        event_type=EventType.TIMER_EVENT,
                        timer_event={
                            "eventId": "string",
                            "name": "lights_off_timeout",
                            "type": "CRON",
                            "time": "2017-09-13T04:18:12.469Z",
                            "expression": "string",
                        },
                    )
                ],
            ),
            settings={
                "property1": "string",
                "property2": "string",
            },
        )
        validate_json_roundtrip(json, expected, LifecycleRequest)
        validate_yaml_roundtrip(None, expected, LifecycleRequest)
        assert expected.token() == "authTokenValue"
        assert expected.app_id() == "d692699d-e7a6-400d-a0b7-d5be96e7a564"
        assert expected.location_id() == "e675a3d9-2499-406c-86dc-8a492a886494"


class TestConvertSmartAppDispatcherConfig:
    def test_config(self):
        expected = SmartAppDispatcherConfig(
            check_signatures=True,
            clock_skew_sec=1234,
            keyserver_url="https://whatever.com",
            log_json=False,
        )
        validate_json_roundtrip(None, expected, SmartAppDispatcherConfig)
        validate_yaml_roundtrip(None, expected, SmartAppDispatcherConfig)


class TestConvertSmartAppDefinition:
    def test_definition(self):
        # SmartAppDefinition isn't part of the SmartThings interface, but we support serializing/deserializing it
        expected = SmartAppDefinition(
            id="id",
            name="name",
            description="description",
            target_url="target_url",
            permissions=["permission"],
            config_pages=[
                SmartAppConfigPage(
                    page_name="First page",
                    sections=[
                        ConfigSection(
                            name="Section 1",
                            settings=[
                                ParagraphSetting(
                                    id="paragraph-id",
                                    name="paragraph-name",
                                    description="paragraph-description",
                                    default_value="paragraph-text",
                                ),
                            ],
                        )
                    ],
                ),
                SmartAppConfigPage(
                    page_name="Second page",
                    sections=[
                        ConfigSection(
                            name="Section 2",
                            settings=[
                                DecimalSetting(
                                    id="decimal-id",
                                    name="decimal-name",
                                    description="decimal-description",
                                    required=False,
                                ),
                            ],
                        )
                    ],
                ),
            ],
        )
        validate_json_roundtrip(None, expected, SmartAppDefinition)
        validate_yaml_roundtrip(None, expected, SmartAppDefinition)



# vim: set ft=python ts=4 sw=4 expandtab:

"""
Verify HTTP signatures on SmartApp lifecycle event requests.
"""

# This implements HTTP signature verification for the SmartApp lifecycle events.
# See: https://developer-preview.smartthings.com/docs/connected-services/hosting/webhook-smartapp/
#
# SmartThings uses Joyent's HTTP signature scheme to sign all lifecycle events.
# See: https://github.com/TritonDataCenter/node-http-signature/blob/master/http_signing.md
#
# Note that only the rsa-sha256 algorithm is supported in this code.  As far as I can
# tell, this is the only one SmartThings uses.  Limiting to the single algorithm greatly
# simplifies the implementation, plus we can use the test cases that Joylent includes in
# their specification document in our own unit tests.
#
# The crypto implementation is PyCryptodome. There are other options, but this appears to
# be well-maintained, is fairly popular (2000+ GitHub stars), and the documentation is good.
# See: https://pypi.org/project/pycryptodomex/ and https://www.pycryptodome.org/en/latest/
#
# The derived signature attributes are in some sense private to the implementation, and in
# some sense public, since they are inputs to the Joyent signature algorithm itself.  I've
# chosen to make them public and read-only on `SignatureVerifier`, which makes the
# implementation more transparent and also makes it easier to verify the implementation
# using the sample data provided in the Joyent spec.
#
# The retrieve_public_key() function is implemented as a function so we can cache the
# result and implement retries outside of SignatureVerifier.  We retry up to 4 times (5
# total attempts), waiting 0.25 seconds before first retry, and limiting the wait between
# retries to 2 seconds.  This is not configurable.  Note that the LRU cache only caches
# responses, not any exceptions that were thrown.

import logging
import re
import urllib.parse
from base64 import b64decode
from collections.abc import Mapping
from functools import lru_cache

import requests
from arrow import Arrow
from arrow import get as arrow_get
from arrow import now as arrow_now
from attrs import field, frozen
from Cryptodome.Hash import SHA256
from Cryptodome.PublicKey import RSA
from Cryptodome.Signature import pkcs1_15
from requests import ConnectionError as RequestsConnectionError
from requests import HTTPError, RequestException
from tenacity import retry
from tenacity.retry import retry_if_exception_type
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_exponential

from smartapp.interface import SignatureError, SmartAppDefinition, SmartAppDispatcherConfig, SmartAppRequestContext

_LOGGER = logging.getLogger(__name__)


@lru_cache(maxsize=32)
@retry(
    stop=stop_after_attempt(5),
    wait=wait_exponential(multiplier=0.25, max=2),
    retry=retry_if_exception_type((RequestsConnectionError, HTTPError)),
)
def retrieve_public_key(key_server_url: str, key_id: str) -> str:
    """Retrieve a public key, caching the result."""
    # Note that the key ID is assumed to be URL-safe per notes in the SmartThings spec, so we don't encode it
    stripped = key_id.lstrip("/")
    url = f"{key_server_url}/{stripped}"
    response = requests.get(url, timeout=5.0)  # excessively long timeout, just in case
    response.raise_for_status()
    return response.text


DATE_FORMAT = "DD MMM YYYY HH:mm:ss ZZZ"  # like "05 Jan 2014 21:31:40 GMT"; we strip off the leading day of week


# noinspection PyUnresolvedReferences
@frozen(kw_only=True, repr=True)
class SignatureVerifier:
    """Signature verifier for Joyent HTTP signatures."""

    context: SmartAppRequestContext = field(repr=False)  # because context.body may contain secrets
    config: SmartAppDispatcherConfig = field()
    definition: SmartAppDefinition = field()
    correlation_id: str | None = field(init=False)
    body: str = field(init=False)
    content_length: int = field(init=False)
    method: str = field(init=False)
    path: str = field(init=False)
    request_target: str = field(init=False)
    date: Arrow = field(init=False)
    authorization: str = field(init=False)
    signing_attributes: Mapping[str, str] = field(init=False)
    signing_headers: str = field(init=False)
    key_id: str = field(init=False)
    keyserver_url: str = field(init=False)
    algorithm: str = field(init=False)
    signature: str = field(init=False)
    signing_string: str = field(init=False)

    @correlation_id.default
    def _default_correlation_id(self) -> str | None:
        return self.context.correlation_id

    @body.default
    def _default_body(self) -> str:
        return self.context.body

    @method.default
    def _default_method(self) -> str:  # noqa: PLR6301
        return "POST"  # this is the only method ever used by the SmartApp interface

    @path.default
    def _default_path(self) -> str:
        # The path from the configured endpoint might be different than what the server served, due to forwarding
        parts = urllib.parse.urlsplit(self.definition.target_url)
        path = parts.path
        if parts.query:
            path = f"{path}?{parts.query}"
        return path

    @request_target.default
    def _default_request_target(self) -> str:
        return f"{self.method.lower()} {self.path}"

    @authorization.default
    def _default_authorization(self) -> str:
        return self.header("authorization")

    # noinspection PyCallingNonCallable
    @date.default
    def _default_date(self) -> Arrow:
        return arrow_get(self.header("date")[5:], DATE_FORMAT)  # remove the day ("Thu, ") from front

    @signing_attributes.default
    def _default_signing_attributes(self) -> Mapping[str, str]:
        # We're parsing a string like: 'Signature keyId="key",algorithm="rsa-sha256",headers="date",signature="xxx"'
        def attribute(name: str, default: str | None = None) -> str:
            if not self.authorization.startswith("Signature "):
                raise SignatureError("Authorization header is not a signature", self.correlation_id)
            pattern = rf"({name}=\")([^\"]+?)(\")"
            match = re.search(pattern=pattern, string=self.authorization)
            if not match:
                if default:
                    return default
                raise SignatureError(f"Signature does not contain: {name}", self.correlation_id)
            return match.group(2)

        return {
            "keyId": attribute("keyId"),
            "headers": attribute("headers", default="Date"),
            "algorithm": attribute("algorithm"),
            "signature": attribute("signature"),
        }

    @signing_headers.default
    def _default_signing_headers(self) -> list[str]:
        # We're parsing a string like "(request-target) date content-type digest" into a list
        return self.signing_attributes["headers"].strip().split(" ")

    @key_id.default
    def _default_key_id(self) -> str:
        return self.signing_attributes["keyId"]

    @keyserver_url.default
    def _default_keyserver_url(self) -> str:
        return self.config.keyserver_url

    @algorithm.default
    def _default_algorithm(self) -> str:
        algorithm = self.signing_attributes["algorithm"]
        if algorithm != "rsa-sha256":
            raise SignatureError(f"Algorithm not supported: {algorithm}", self.correlation_id)
        return algorithm

    @signature.default
    def _default_signature(self) -> str:
        return self.signing_attributes["signature"]

    @signing_string.default
    def _signing_string(self) -> str:
        # The only "special" component in the signing string is "(request-target)"
        # Everything else must be found as a header, and we should fail if it's not
        components = []
        for name in self.signing_headers:
            if name == "(request-target)":
                components.append(f"(request-target): {self.request_target}")
            else:
                components.append(f"{name.lower()}: {self.header(name)}")
        return "\n".join(components).rstrip("\n")

    def header(self, name: str) -> str:
        """Return the named header case-insensitively, raising SignatureError if it is not found or is empty"""
        value = self.context.header(name)
        if not value:
            raise SignatureError(f"Header not found: {name}", self.correlation_id)
        return value

    def retrieve_public_key(self) -> str:
        """Retrieve the configured public key."""
        try:
            return retrieve_public_key(self.keyserver_url, self.key_id)  # will retry automatically
        except RequestException as e:
            raise SignatureError(f"Failed to retrieve key [{self.key_id}]", self.correlation_id) from e

    def verify_date(self) -> None:
        """Verify the date, ensuring that it is current per skew configuration."""
        if self.config.clock_skew_sec is not None:
            skew = abs(arrow_now() - self.date)
            if skew.seconds > self.config.clock_skew_sec:
                raise SignatureError(f"Request date is not current, skew of {skew.seconds} seconds", self.correlation_id)

    def verify_signature(self) -> None:
        """Verify the RSA-SHA256 signature of the signing string."""
        # See: https://www.pycryptodome.org/en/latest/src/signature/pkcs1_v1_5.html
        try:
            _LOGGER.debug("[%s] Signing string: \n%s", self.correlation_id, self.signing_string)
            signature = b64decode(self.signature)
            sha256 = SHA256.new(self.signing_string.encode())
            key = RSA.import_key(self.retrieve_public_key())
            pkcs1_15.new(key).verify(sha256, signature)
        except (ValueError, TypeError) as e:
            raise SignatureError("Signature is not valid", self.correlation_id) from e

    def verify(self) -> None:
        """Verify the request date and the signature of the signing string."""
        self.verify_date()
        self.verify_signature()



# vim: set ft=python ts=4 sw=4 expandtab:

"""
Classes that are part of the SmartApp interface.
"""

# For lifecycle class definitions, see:
#
#   https://developer-preview.smartthings.com/docs/connected-services/lifecycles/
#   https://developer-preview.smartthings.com/docs/connected-services/configuration/
#
# There is not any public documentation about event structure, only the Javascript
# reference implementation here:
#
#   https://github.com/SmartThingsCommunity/smartapp-sdk-nodejs/blob/f1ef97ec9c6dc270ba744197b842c6632c778987/lib/lifecycle-events.d.ts
#
# However, as of this writitng, even that reference implementation is not fully up-to-date
# with the JSON that is being returned for some events I have examined in my testing.
#
# I have access to private documentation that shows all of the attributes.  However, that
# documentation doesn't always make it clear which attributes will always be included and
# which are optional.  As compromise, I have decided to maintain the actual events as
# dicts rather than true objects.  See further discussion below by the Event class.

from abc import ABC, abstractmethod
from collections.abc import Callable, Mapping
from enum import Enum, StrEnum
from typing import Any, assert_never

from arrow import Arrow
from attrs import field, frozen

AUTHORIZATION_HEADER = "authorization"
CORRELATION_ID_HEADER = "x-st-correlation"
DATE_HEADER = "date"


class LifecyclePhase(Enum):
    """Lifecycle phases."""

    CONFIRMATION = "CONFIRMATION"
    CONFIGURATION = "CONFIGURATION"
    INSTALL = "INSTALL"
    UPDATE = "UPDATE"
    UNINSTALL = "UNINSTALL"
    OAUTH_CALLBACK = "OAUTH_CALLBACK"
    EVENT = "EVENT"


class ConfigValueType(Enum):
    """Types of config values."""

    DEVICE = "DEVICE"
    STRING = "STRING"


class ConfigPhase(Enum):
    """Sub-phases within the CONFIGURATION phase."""

    INITIALIZE = "INITIALIZE"
    PAGE = "PAGE"


class ConfigSettingType(Enum):
    """Types of config settings."""

    DEVICE = "DEVICE"
    TEXT = "TEXT"
    BOOLEAN = "BOOLEAN"
    ENUM = "ENUM"
    LINK = "LINK"
    PAGE = "PAGE"
    IMAGE = "IMAGE"
    ICON = "ICON"
    TIME = "TIME"
    PARAGRAPH = "PARAGRAPH"
    EMAIL = "EMAIL"
    DECIMAL = "DECIMAL"
    NUMBER = "NUMBER"
    PHONE = "PHONE"
    OAUTH = "OAUTH"


class EventType(Enum):
    """Supported event types."""

    DEVICE_COMMANDS_EVENT = "DEVICE_COMMANDS_EVENT"
    DEVICE_EVENT = "DEVICE_EVENT"
    DEVICE_HEALTH_EVENT = "DEVICE_HEALTH_EVENT"
    DEVICE_LIFECYCLE_EVENT = "DEVICE_LIFECYCLE_EVENT"
    HUB_HEALTH_EVENT = "HUB_HEALTH_EVENT"
    INSTALLED_APP_LIFECYCLE_EVENT = "INSTALLED_APP_LIFECYCLE_EVENT"
    MODE_EVENT = "MODE_EVENT"
    SCENE_LIFECYCLE_EVENT = "SCENE_LIFECYCLE_EVENT"
    SECURITY_ARM_STATE_EVENT = "SECURITY_ARM_STATE_EVENT"
    TIMER_EVENT = "TIMER_EVENT"
    WEATHER_EVENT = "WEATHER_EVENT"


class SubscriptionType(Enum):
    """Supported subscription types."""

    DEVICE = "DEVICE"
    CAPABILITY = "CAPABILITY"
    MODE = "MODE"
    DEVICE_LIFECYCLE = "DEVICE_LIFECYCLE"
    DEVICE_HEALTH = "DEVICE_HEALTH"
    SECURITY_ARM_STATE = "SECURITY_ARM_STATE"
    HUB_HEALTH = "HUB_HEALTH"
    SCENE_LIFECYCLE = "SCENE_LIFECYCLE"


class BooleanValue(StrEnum):
    """String boolean values."""

    TRUE = "true"
    FALSE = "false"


@frozen(kw_only=True)
class AbstractRequest(ABC):
    """Abstract parent class for all types of lifecycle requests."""

    lifecycle: LifecyclePhase
    execution_id: str
    locale: str
    version: str


@frozen(kw_only=True)
class AbstractSetting(ABC):
    """Abstract parent class for all types of config settings."""

    id: str
    name: str
    description: str
    required: bool | None = False


@frozen(kw_only=True)
class DeviceSetting(AbstractSetting):
    """A DEVICE setting."""

    type: ConfigSettingType = ConfigSettingType.DEVICE
    multiple: bool
    capabilities: list[str]  # note that this is treated as AND - you'll get devices that have all capabilities
    permissions: list[str]


@frozen(kw_only=True)
class TextSetting(AbstractSetting):
    """A TEXT setting."""

    type: ConfigSettingType = ConfigSettingType.TEXT
    default_value: str


@frozen(kw_only=True)
class BooleanSetting(AbstractSetting):
    """A BOOLEAN setting."""

    type: ConfigSettingType = ConfigSettingType.BOOLEAN
    default_value: BooleanValue


@frozen(kw_only=True)
class EnumOption:
    """An option within an ENUM setting"""

    id: str
    name: str


@frozen(kw_only=True)
class EnumOptionGroup:
    """A group of options within an ENUM setting"""

    name: str
    options: list[EnumOption]


@frozen(kw_only=True)
class EnumSetting(AbstractSetting):
    """An ENUM setting."""

    type: ConfigSettingType = ConfigSettingType.ENUM
    multiple: bool
    options: list[EnumOption] | None = None
    grouped_options: list[EnumOptionGroup] | None = None


@frozen(kw_only=True)
class LinkSetting(AbstractSetting):
    """A LINK setting."""

    type: ConfigSettingType = ConfigSettingType.LINK
    url: str
    image: str


@frozen(kw_only=True)
class PageSetting(AbstractSetting):
    """A PAGE setting."""

    type: ConfigSettingType = ConfigSettingType.PAGE
    page: str
    image: str


@frozen(kw_only=True)
class ImageSetting(AbstractSetting):
    """An IMAGE setting."""

    type: ConfigSettingType = ConfigSettingType.IMAGE
    image: str


@frozen(kw_only=True)
class IconSetting(AbstractSetting):
    """An ICON setting."""

    type: ConfigSettingType = ConfigSettingType.ICON
    image: str


@frozen(kw_only=True)
class TimeSetting(AbstractSetting):
    """A TIME setting."""

    type: ConfigSettingType = ConfigSettingType.TIME


@frozen(kw_only=True)
class ParagraphSetting(AbstractSetting):
    """A PARAGRAPH setting."""

    type: ConfigSettingType = ConfigSettingType.PARAGRAPH
    default_value: str


@frozen(kw_only=True)
class EmailSetting(AbstractSetting):
    """An EMAIL setting."""

    type: ConfigSettingType = ConfigSettingType.EMAIL


@frozen(kw_only=True)
class DecimalSetting(AbstractSetting):
    """A DECIMAL setting."""

    type: ConfigSettingType = ConfigSettingType.DECIMAL


@frozen(kw_only=True)
class NumberSetting(AbstractSetting):
    """A NUMBER setting."""

    type: ConfigSettingType = ConfigSettingType.NUMBER


@frozen(kw_only=True)
class PhoneSetting(AbstractSetting):
    """A PHONE setting."""

    type: ConfigSettingType = ConfigSettingType.PHONE


@frozen(kw_only=True)
class OauthSetting(AbstractSetting):
    """An OAUTH setting."""

    type: ConfigSettingType = ConfigSettingType.OAUTH
    browser: bool
    url_template: str


ConfigSetting = (
    DeviceSetting
    | TextSetting
    | BooleanSetting
    | EnumSetting
    | LinkSetting
    | PageSetting
    | ImageSetting
    | IconSetting
    | TimeSetting
    | ParagraphSetting
    | EmailSetting
    | DecimalSetting
    | NumberSetting
    | PhoneSetting
    | OauthSetting
)


@frozen(kw_only=True)
class DeviceValue:
    device_id: str
    component_id: str


@frozen(kw_only=True)
class DeviceConfigValue:
    """DEVICE configuration value."""

    device_config: DeviceValue
    value_type: ConfigValueType = ConfigValueType.DEVICE


@frozen(kw_only=True)
class StringValue:
    value: str


@frozen(kw_only=True)
class StringConfigValue:
    """STRING configuration value."""

    string_config: StringValue
    value_type: ConfigValueType = ConfigValueType.STRING


ConfigValue = DeviceConfigValue | StringConfigValue


@frozen(kw_only=True)
class InstalledApp:
    """Installed application."""

    installed_app_id: str
    location_id: str
    config: dict[str, list[ConfigValue]]
    permissions: list[str] = field(factory=list)

    def as_devices(self, key: str) -> list[DeviceValue]:
        """Return a list of devices for a named configuration value."""
        return [item.device_config for item in self.config[key]]  # type: ignore[union-attr]

    def as_str(self, key: str) -> str:
        """Return a named configuration value, interpreted as a string"""
        return self.config[key][0].string_config.value  # type: ignore[union-attr]

    def as_bool(self, key: str) -> bool:
        """Return a named configuration value, interpreted as a boolean"""
        return bool(self.as_str(key))

    def as_int(self, key: str) -> int:
        """Return a named configuration value, interpreted as an integer"""
        return int(self.as_str(key))

    def as_float(self, key: str) -> float:
        """Return a named configuration value, interpreted as a float"""
        return float(self.as_str(key))


@frozen(kw_only=True)
class Event:
    """Holds the triggered event, one of several different attributes depending on event type."""

    event_time: Arrow | None = None
    event_type: EventType
    device_event: dict[str, Any] | None = None
    device_lifecycle_event: dict[str, Any] | None = None
    device_health_event: dict[str, Any] | None = None
    device_commands_event: dict[str, Any] | None = None
    mode_event: dict[str, Any] | None = None
    timer_event: dict[str, Any] | None = None
    scene_lifecycle_event: dict[str, Any] | None = None
    security_arm_state_event: dict[str, Any] | None = None
    hub_health_event: dict[str, Any] | None = None
    installed_app_lifecycle_event: dict[str, Any] | None = None
    weather_event: dict[str, Any] | None = None
    weather_data: dict[str, Any] | None = None
    air_quality_data: dict[str, Any] | None = None

    # noinspection PyUnreachableCode
    def for_type(self, event_type: EventType) -> dict[str, Any] | None:  # noqa: PLR0911
        """Return the attribute associated with an event type."""
        if event_type == EventType.DEVICE_COMMANDS_EVENT:
            return self.device_commands_event
        if event_type == EventType.DEVICE_EVENT:
            return self.device_event
        if event_type == EventType.DEVICE_HEALTH_EVENT:
            return self.device_health_event
        if event_type == EventType.DEVICE_LIFECYCLE_EVENT:
            return self.device_lifecycle_event
        if event_type == EventType.HUB_HEALTH_EVENT:
            return self.hub_health_event
        if event_type == EventType.INSTALLED_APP_LIFECYCLE_EVENT:
            return self.installed_app_lifecycle_event
        if event_type == EventType.MODE_EVENT:
            return self.mode_event
        if event_type == EventType.SCENE_LIFECYCLE_EVENT:
            return self.scene_lifecycle_event
        if event_type == EventType.SECURITY_ARM_STATE_EVENT:
            return self.security_arm_state_event
        if event_type == EventType.TIMER_EVENT:
            return self.timer_event
        if event_type == EventType.WEATHER_EVENT:
            return self.weather_event
        assert_never(event_type)


@frozen(kw_only=True)
class ConfirmationData:
    """Confirmation data."""

    app_id: str
    confirmation_url: str


@frozen(kw_only=True)
class ConfigInit:
    """Initialization data."""

    id: str
    name: str
    description: str
    permissions: list[str]
    first_page_id: str


@frozen(kw_only=True)
class ConfigRequestData:
    """Configuration data provided on the request."""

    installed_app_id: str
    phase: ConfigPhase
    page_id: str
    previous_page_id: str
    config: dict[str, list[ConfigValue]]


@frozen(kw_only=True)
class ConfigInitData:
    """Configuration data provided in an INITIALIZATION response."""

    initialize: ConfigInit


@frozen(kw_only=True)
class ConfigSection:
    """A section within a configuration page."""

    name: str
    settings: list[ConfigSetting]


@frozen(kw_only=True)
class ConfigPage:
    """A page of configuration data for the CONFIGURATION phase."""

    page_id: str
    name: str
    previous_page_id: str | None
    next_page_id: str | None
    complete: bool
    sections: list[ConfigSection]


@frozen(kw_only=True)
class ConfigPageData:
    """Configuration data provided in an PAGE response."""

    page: ConfigPage


@frozen(kw_only=True)
class InstallData:
    """Install data."""

    # note: auth_token and refresh_token are secrets, so we don't include them in string output

    auth_token: str = field(repr=False)
    refresh_token: str = field(repr=False)
    installed_app: InstalledApp

    def token(self) -> str:
        """Return the auth token associated with this request."""
        return self.auth_token

    def app_id(self) -> str:
        """Return the installed application id associated with this request."""
        return self.installed_app.installed_app_id

    def location_id(self) -> str:
        """Return the installed location id associated with this request."""
        return self.installed_app.location_id

    def as_devices(self, key: str) -> list[DeviceValue]:
        """Return a list of devices for a named configuration value."""
        return self.installed_app.as_devices(key)

    def as_str(self, key: str) -> str:
        """Return a named configuration value, interpreted as a string"""
        return self.installed_app.as_str(key)

    def as_bool(self, key: str) -> bool:
        """Return a named configuration value, interpreted as a boolean"""
        return self.installed_app.as_bool(key)

    def as_int(self, key: str) -> int:
        """Return a named configuration value, interpreted as an integer"""
        return self.installed_app.as_int(key)

    def as_float(self, key: str) -> float:
        """Return a named configuration value, interpreted as a float"""
        return self.installed_app.as_float(key)


@frozen(kw_only=True)
class UpdateData:
    """Update data."""

    # note: auth_token and refresh_token are secrets, so we don't include them in string output

    auth_token: str = field(repr=False)
    refresh_token: str = field(repr=False)
    installed_app: InstalledApp
    previous_config: dict[str, list[ConfigValue]] | None = None
    previous_permissions: list[str] = field(factory=list)

    def token(self) -> str:
        """Return the auth token associated with this request."""
        return self.auth_token

    def app_id(self) -> str:
        """Return the installed application id associated with this request."""
        return self.installed_app.installed_app_id

    def location_id(self) -> str:
        """Return the installed location id associated with this request."""
        return self.installed_app.location_id

    def as_devices(self, key: str) -> list[DeviceValue]:
        """Return a list of devices for a named configuration value."""
        return self.installed_app.as_devices(key)

    def as_str(self, key: str) -> str:
        """Return a named configuration value, interpreted as a string"""
        return self.installed_app.as_str(key)

    def as_bool(self, key: str) -> bool:
        """Return a named configuration value, interpreted as a boolean"""
        return self.installed_app.as_bool(key)

    def as_int(self, key: str) -> int:
        """Return a named configuration value, interpreted as an integer"""
        return self.installed_app.as_int(key)

    def as_float(self, key: str) -> float:
        """Return a named configuration value, interpreted as a float"""
        return self.installed_app.as_float(key)


@frozen(kw_only=True)
class UninstallData:
    """Install data."""

    installed_app: InstalledApp

    def app_id(self) -> str:
        """Return the installed application id associated with this request."""
        return self.installed_app.installed_app_id

    def location_id(self) -> str:
        """Return the installed location id associated with this request."""
        return self.installed_app.location_id


@frozen(kw_only=True)
class OauthCallbackData:
    installed_app_id: str
    url_path: str


@frozen(kw_only=True)
class EventData:
    """Event data."""

    # note: auth_token is a secret, so we don't include it in string output

    auth_token: str = field(repr=False)
    installed_app: InstalledApp
    events: list[Event]

    def token(self) -> str:
        """Return the auth token associated with this request."""
        return self.auth_token

    def app_id(self) -> str:
        """Return the installed application id associated with this request."""
        return self.installed_app.installed_app_id

    def location_id(self) -> str:
        """Return the installed location id associated with this request."""
        return self.installed_app.location_id

    def for_type(self, event_type: EventType) -> list[dict[str, Any]]:
        """Get all events for a particular event type, possibly empty."""
        return [
            event.for_type(event_type)  # type: ignore[misc]
            for event in self.events
            if event.event_type == event_type and event.for_type(event_type) is not None
        ]

    def filter(self, event_type: EventType, predicate: Callable[[dict[str, Any]], bool] | None = None) -> list[dict[str, Any]]:
        """Apply a filter to a set of events with a particular event type."""
        return list(filter(predicate, self.for_type(event_type)))


@frozen(kw_only=True)
class ConfirmationRequest(AbstractRequest):
    """Request for CONFIRMATION phase"""

    app_id: str
    confirmation_data: ConfirmationData
    settings: dict[str, Any] = field(factory=dict)


@frozen(kw_only=True)
class ConfirmationResponse:
    """Response for CONFIRMATION phase"""

    target_url: str


@frozen(kw_only=True)
class ConfigurationRequest(AbstractRequest):
    """Request for CONFIGURATION phase"""

    configuration_data: ConfigRequestData
    settings: dict[str, Any] = field(factory=dict)


@frozen(kw_only=True)
class ConfigurationInitResponse:
    """Response for CONFIGURATION/INITIALIZE phase"""

    configuration_data: ConfigInitData


@frozen(kw_only=True)
class ConfigurationPageResponse:
    """Response for CONFIGURATION/PAGE phase"""

    configuration_data: ConfigPageData


@frozen(kw_only=True)
class InstallRequest(AbstractRequest):
    """Request for INSTALL phase"""

    install_data: InstallData
    settings: dict[str, Any] = field(factory=dict)

    def token(self) -> str:
        """Return the auth token associated with this request."""
        return self.install_data.token()

    def app_id(self) -> str:
        """Return the installed application id associated with this request."""
        return self.install_data.app_id()

    def location_id(self) -> str:
        """Return the installed location id associated with this request."""
        return self.install_data.location_id()

    def as_devices(self, key: str) -> list[DeviceValue]:
        """Return a list of devices for a named configuration value."""
        return self.install_data.as_devices(key)

    def as_str(self, key: str) -> str:
        """Return a named configuration value, interpreted as a string"""
        return self.install_data.as_str(key)

    def as_bool(self, key: str) -> bool:
        """Return a named configuration value, interpreted as a boolean"""
        return self.install_data.as_bool(key)

    def as_int(self, key: str) -> int:
        """Return a named configuration value, interpreted as an integer"""
        return self.install_data.as_int(key)

    def as_float(self, key: str) -> float:
        """Return a named configuration value, interpreted as a float"""
        return self.install_data.as_float(key)


@frozen(kw_only=True)
class InstallResponse:
    """Response for INSTALL phase"""

    install_data: dict[str, Any] = field(factory=dict)  # always empty in the response


@frozen(kw_only=True)
class UpdateRequest(AbstractRequest):
    """Request for UPDATE phase"""

    update_data: UpdateData
    settings: dict[str, Any] = field(factory=dict)

    def token(self) -> str:
        """Return the auth token associated with this request."""
        return self.update_data.token()

    def app_id(self) -> str:
        """Return the installed application id associated with this request."""
        return self.update_data.app_id()

    def location_id(self) -> str:
        """Return the installed location id associated with this request."""
        return self.update_data.location_id()

    def as_devices(self, key: str) -> list[DeviceValue]:
        """Return a list of devices for a named configuration value."""
        return self.update_data.as_devices(key)

    def as_str(self, key: str) -> str:
        """Return a named configuration value, interpreted as a string"""
        return self.update_data.as_str(key)

    def as_bool(self, key: str) -> bool:
        """Return a named configuration value, interpreted as a boolean"""
        return self.update_data.as_bool(key)

    def as_int(self, key: str) -> int:
        """Return a named configuration value, interpreted as an integer"""
        return self.update_data.as_int(key)

    def as_float(self, key: str) -> float:
        """Return a named configuration value, interpreted as a float"""
        return self.update_data.as_float(key)


@frozen(kw_only=True)
class UpdateResponse:
    """Response for UPDATE phase"""

    update_data: dict[str, Any] = field(factory=dict)  # always empty in the response


@frozen(kw_only=True)
class UninstallRequest(AbstractRequest):
    """Request for UNINSTALL phase"""

    uninstall_data: UninstallData
    settings: dict[str, Any] = field(factory=dict)

    def app_id(self) -> str:
        """Return the installed application id associated with this request."""
        return self.uninstall_data.app_id()

    def location_id(self) -> str:
        """Return the installed location id associated with this request."""
        return self.uninstall_data.location_id()


@frozen(kw_only=True)
class UninstallResponse:
    """Response for UNINSTALL phase"""

    uninstall_data: dict[str, Any] = field(factory=dict)  # always empty in the response


@frozen(kw_only=True)
class OauthCallbackRequest(AbstractRequest):
    """Request for OAUTH_CALLBACK phase"""

    o_auth_callback_data: OauthCallbackData


@frozen(kw_only=True)
class OauthCallbackResponse:
    """Response for OAUTH_CALLBACK phase"""

    o_auth_callback_data: dict[str, Any] = field(factory=dict)  # always empty in the response


@frozen(kw_only=True)
class EventRequest(AbstractRequest):
    """Request for EVENT phase"""

    event_data: EventData
    settings: dict[str, Any] = field(factory=dict)

    def token(self) -> str:
        """Return the auth token associated with this request."""
        return self.event_data.token()

    def app_id(self) -> str:
        """Return the installed application id associated with this request."""
        return self.event_data.app_id()

    def location_id(self) -> str:
        """Return the installed location id associated with this request."""
        return self.event_data.location_id()


@frozen(kw_only=True)
class EventResponse:
    """Response for EVENT phase"""

    event_data: dict[str, Any] = field(factory=dict)  # always empty in the response


LifecycleRequest = (
    ConfigurationRequest
    | ConfirmationRequest
    | InstallRequest
    | UpdateRequest
    | UninstallRequest
    | OauthCallbackRequest
    | EventRequest
)


LifecycleResponse = (
    ConfigurationInitResponse
    | ConfigurationPageResponse
    | ConfirmationResponse
    | InstallResponse
    | UpdateResponse
    | UninstallResponse
    | OauthCallbackResponse
    | EventResponse
)


REQUEST_BY_PHASE = {
    LifecyclePhase.CONFIGURATION: ConfigurationRequest,
    LifecyclePhase.CONFIRMATION: ConfirmationRequest,
    LifecyclePhase.INSTALL: InstallRequest,
    LifecyclePhase.UPDATE: UpdateRequest,
    LifecyclePhase.UNINSTALL: UninstallRequest,
    LifecyclePhase.OAUTH_CALLBACK: OauthCallbackRequest,
    LifecyclePhase.EVENT: EventRequest,
}

CONFIG_VALUE_BY_TYPE = {
    ConfigValueType.DEVICE: DeviceConfigValue,
    ConfigValueType.STRING: StringConfigValue,
}

CONFIG_SETTING_BY_TYPE = {
    ConfigSettingType.DEVICE: DeviceSetting,
    ConfigSettingType.TEXT: TextSetting,
    ConfigSettingType.BOOLEAN: BooleanSetting,
    ConfigSettingType.ENUM: EnumSetting,
    ConfigSettingType.LINK: LinkSetting,
    ConfigSettingType.PAGE: PageSetting,
    ConfigSettingType.IMAGE: ImageSetting,
    ConfigSettingType.ICON: IconSetting,
    ConfigSettingType.TIME: TimeSetting,
    ConfigSettingType.PARAGRAPH: ParagraphSetting,
    ConfigSettingType.EMAIL: EmailSetting,
    ConfigSettingType.DECIMAL: DecimalSetting,
    ConfigSettingType.NUMBER: NumberSetting,
    ConfigSettingType.PHONE: PhoneSetting,
    ConfigSettingType.OAUTH: OauthSetting,
}


@frozen
class SmartAppError(Exception):
    """An error tied to the SmartApp implementation."""

    message: str
    correlation_id: str | None = None


@frozen
class InternalError(SmartAppError):
    """An internal error was encountered processing a lifecycle event."""


@frozen
class BadRequestError(SmartAppError):
    """A lifecycle event was invalid."""


@frozen
class SignatureError(SmartAppError):
    """The request signature on a lifecycle event was invalid."""


@frozen(kw_only=True)
class SmartAppDispatcherConfig:
    # noinspection PyUnresolvedReferences
    """
    Configuration for the SmartAppDispatcher.

    Any production SmartApp should always check signatures.  We support disabling that feature
    to make local testing easier during development.

    BEWARE: setting `log_json` to `True` will potentially place secrets (such as authorization
    keys) in your logs.  This is intended for use during development and debugging only.

    Attributes:
        check_signatures(bool): Whether to check the digital signature on lifecycle requests
        clock_skew_sec(int): Amount of clock skew allowed when verifying digital signatures, or None to allow any skew
        keyserver_url(str): The SmartThings keyserver URL, where we retrieve keys for signature checks
        log_json(bool): Whether to log JSON data at DEBUG level when processing requests
    """

    check_signatures: bool = True
    clock_skew_sec: int | None = 300
    keyserver_url: str = "https://key.smartthings.com"
    log_json: bool = False


class SmartAppEventHandler(ABC):
    """
    Application event handler for SmartApp lifecycle events.

    Inherit from this class to implement your own application-specific event handler.
    The application-specific event handler is always called first, before any default
    event handler logic in the dispatcher itself.

    The correlation id is an optional value that you can associate with your log messages.
    It may aid in debugging if you need to contact SmartThings for support.

    Some lifecycle events do not require you to implement any custom event handler logic:

    - CONFIRMATION: normally no callback needed, since the dispatcher logs the app id and confirmation URL
    - CONFIGURATION: normally no callback needed, since the dispatcher has the information it needs to respond
    - INSTALL/UPDATE: set up or replace subscriptions and schedules and persist required data, if any
    - UNINSTALL: remove persisted data, if any
    - OAUTH_CALLBACK: coordinate with your oauth provider as needed
    - EVENT: handle SmartThings events or scheduled triggers

    The EventRequest object that you receive for the EVENT callback includes an
    authorization token and also the entire configuration bundle for the installed
    application.  So, if your SmartApp is built around event handling and scheduled
    actions triggered by SmartThings, your handler can probably be stateless.  There is
    probably is not any need to persist any of the data returned in the INSTALL or UPDATE
    lifecycle events into your own data store.

    Note that SmartAppHandler is a synchronous and single-threaded interface.  The
    assumption is that if you need high-volume asynchronous or multi-threaded processing,
    you will implement that at the tier above this where the actual POST requests are
    accepted from remote callers.
    """

    @abstractmethod
    def handle_confirmation(self, correlation_id: str | None, request: ConfirmationRequest) -> None:
        """Handle a CONFIRMATION lifecycle request"""

    @abstractmethod
    def handle_configuration(self, correlation_id: str | None, request: ConfigurationRequest) -> None:
        """Handle a CONFIGURATION lifecycle request."""

    @abstractmethod
    def handle_install(self, correlation_id: str | None, request: InstallRequest) -> None:
        """Handle an INSTALL lifecycle request."""

    @abstractmethod
    def handle_update(self, correlation_id: str | None, request: UpdateRequest) -> None:
        """Handle an UPDATE lifecycle request."""

    @abstractmethod
    def handle_uninstall(self, correlation_id: str | None, request: UninstallRequest) -> None:
        """Handle an UNINSTALL lifecycle request."""

    @abstractmethod
    def handle_oauth_callback(self, correlation_id: str | None, request: OauthCallbackRequest) -> None:
        """Handle an OAUTH_CALLBACK lifecycle request."""

    @abstractmethod
    def handle_event(self, correlation_id: str | None, request: EventRequest) -> None:
        """Handle an EVENT lifecycle request."""


@frozen(kw_only=True)
class SmartAppConfigPage:
    """
    A page of configuration for the SmartApp.
    """

    page_name: str
    sections: list[ConfigSection]


@frozen(kw_only=True)
class SmartAppDefinition:
    # noinspection PyUnresolvedReferences
    """
    The definition of the SmartApp.

    All of this data would normally be static for any given version of your application.
    If you wish, you can maintain the definition in YAML or JSON in your source tree
    and parse it with `smartapp.converter.CONVERTER`.

    Keep in mind that the JSON or YAML format on disk will be consistent with the SmartThings
    lifecycle API, so it will use camel case attribute names (like `configPages`) rather than
    the Python attribute names you see in source code (like `config_pages`).

    Attributes:
        id(str): Identifier for this SmartApp
        name(str): Name of the SmartApp
        description(str): Description of the SmartApp
        permissions(List[str]): Permissions that the SmartApp requires
        config_pages(List[SmartAppConfigPage]): Configuration pages that the SmartApp will offer users
    """

    id: str
    name: str
    description: str
    target_url: str
    permissions: list[str]
    config_pages: list[SmartAppConfigPage] | None


# noinspection PyShadowingBuiltins,PyMethodMayBeStatic
class SmartAppConfigManager(ABC):
    """
    Configuration manager, used by the dispatcher to respond to CONFIGURATION events.

    The dispatcher has a default configuration manager.  However, you can implement your
    own if that default behavior does not meet your needs.  For instance, a static config
    definition is adequate for lots of SmartApps, but it doesn't work for some types of
    complex configuration, where the responses need to be generated dynamically.  In that
    case, you can implement your own configuration manager with that specialized behavior.

    This abstract class also includes several convenience methods to make it easier to
    build responses.
    """

    def handle_initialize(self, _request: ConfigurationRequest, definition: SmartAppDefinition) -> ConfigurationInitResponse:
        """Handle a CONFIGURATION INITIALIZE lifecycle request."""
        return self.build_init_response(
            id=definition.id,
            name=definition.name,
            description=definition.description,
            permissions=definition.permissions,
            first_page_id=1,
        )

    @abstractmethod
    def handle_page(self, request: ConfigurationRequest, definition: SmartAppDefinition, page_id: int) -> ConfigurationPageResponse:
        """Handle a CONFIGURATION PAGE lifecycle request."""

    def build_init_response(  # noqa: PLR6301
        self,
        id: str,  # noqa: A002
        name: str,
        description: str,
        permissions: list[str],
        first_page_id: int,
    ) -> ConfigurationInitResponse:
        """Build a ConfigurationInitResponse."""
        return ConfigurationInitResponse(
            configuration_data=ConfigInitData(
                initialize=ConfigInit(
                    id=id,
                    name=name,
                    description=description,
                    permissions=permissions,
                    first_page_id=str(first_page_id),
                )
            )
        )

    def build_page_response(  # noqa: PLR0913,PLR0917,PLR6301
        self,
        page_id: int,
        name: str,
        previous_page_id: int | None,
        next_page_id: int | None,
        complete: bool,  # noqa: FBT001
        sections: list[ConfigSection],
    ) -> ConfigurationPageResponse:
        """Build a ConfigurationPageResponse."""
        return ConfigurationPageResponse(
            configuration_data=ConfigPageData(
                page=ConfigPage(
                    name=name,
                    page_id=str(page_id),
                    previous_page_id=str(previous_page_id) if previous_page_id else None,
                    next_page_id=str(next_page_id) if next_page_id else None,
                    complete=complete,
                    sections=sections,
                )
            )
        )


# noinspection PyUnresolvedReferences
@frozen(kw_only=True)
class SmartAppRequestContext:
    """
    The context for a SmartApp lifecycle request.

    Attributes:
        headers(Mapping[str, str]): The request headers
        body(str): The body of the request as string
    """

    # I'm pulling out the correlation id, signature, and date because they are 3 specific
    # headers that I know the SmartThings API always provides.  Others can be pulled out
    # using header().

    headers: Mapping[str, str] = field(factory=dict)
    body: str = ""
    normalized: Mapping[str, str] = field(init=False)
    correlation_id: str = field(init=False)
    signature: str = field(init=False)
    date: str = field(init=False)

    @normalized.default
    def _default_normalized(self) -> Mapping[str, str]:
        # in conjunction with header(), this gives us a case-insensitive dictionary
        return {key.lower(): value for (key, value) in self.headers.items()} if self.headers else {}

    @correlation_id.default
    def _default_correlation_id(self) -> str | None:
        return self.header(CORRELATION_ID_HEADER)

    @signature.default
    def _default_signature(self) -> str | None:
        return self.header(AUTHORIZATION_HEADER)

    @date.default
    def _default_date(self) -> str | None:
        return self.header(DATE_HEADER)

    def header(self, name: str) -> str | None:
        """Return the named header case-insensitively, or None if not found."""
        if name.lower() not in self.normalized:
            return None
        value = self.normalized[name.lower()]
        if not value or not value.strip():
            return None
        return value



# vim: set ft=python ts=4 sw=4 expandtab:
from pathlib import Path

import pytest

from smartapp.converter import CONVERTER
from smartapp.interface import (
    BadRequestError,
    DeviceValue,
    Event,
    EventRequest,
    EventType,
    InstallRequest,
    InternalError,
    SignatureError,
    SmartAppError,
    SmartAppRequestContext,
    UpdateRequest,
)
from tests.smartapp.testutil import load_file

FIXTURE_DIR = Path(__file__).parent / "fixtures"

DEVICE_EVENT = {
    "subscriptionName": "motion_sensors",
    "eventId": "736e3903-001c-4d40-b408-ff40d162a06b",
    "locationId": "499e28ba-b33b-49c9-a5a1-cce40e41f8a6",
    "deviceId": "6f5ea629-4c05-4a90-a244-cc129b0a80c3",
    "componentId": "main",
    "capability": "motionSensor",
    "attribute": "motion",
    "value": "active",
    "stateChange": True,
}

TIMER_EVENT = {
    "eventId": "string",
    "name": "lights_off_timeout",
    "type": "CRON",
    "time": "2017-09-13T04:18:12.469Z",
    "expression": "string",
}


class TestExceptions:
    @pytest.mark.parametrize(
        "exception",
        [
            SmartAppError,
            InternalError,
            BadRequestError,
            SignatureError,
        ],
    )
    def test_exceptions(self, exception):
        e = exception("message")
        assert e.message == "message"
        assert e.correlation_id is None
        assert isinstance(e, SmartAppError)
        e = exception("message", "id")
        assert e.message == "message"
        assert e.correlation_id == "id"
        assert isinstance(e, SmartAppError)


class TestSmartAppRequestContext:
    def test_context(self):
        headers = {
            "Date": "the-date",
            "Authorization": "signature",
            "X-ST-Correlation": "correlation",
            "empty": "",
            "whitespace": "\t\n",
            "none": None,
        }
        context = SmartAppRequestContext(
            headers=headers,
            body="thebody",
        )
        assert context.correlation_id == "correlation"
        assert context.signature == "signature"
        assert context.date == "the-date"
        for header in ["DATE", "Date", "date"]:
            assert context.header(header) == "the-date"
        for header in ["AUTHORIZATION", "Authorization", "authorization"]:
            assert context.header(header) == "signature"
        for header in ["missing", "empty", "whitespace", "none"]:
            assert context.header(header) is None


class TestEvent:
    @pytest.mark.parametrize(
        "event_type,attribute",
        [
            [EventType.DEVICE_COMMANDS_EVENT, "device_commands_event"],
            [EventType.DEVICE_EVENT, "device_event"],
            [EventType.DEVICE_HEALTH_EVENT, "device_health_event"],
            [EventType.DEVICE_LIFECYCLE_EVENT, "device_lifecycle_event"],
            [EventType.HUB_HEALTH_EVENT, "hub_health_event"],
            [EventType.INSTALLED_APP_LIFECYCLE_EVENT, "installed_app_lifecycle_event"],
            [EventType.MODE_EVENT, "mode_event"],
            [EventType.SCENE_LIFECYCLE_EVENT, "scene_lifecycle_event"],
            [EventType.SECURITY_ARM_STATE_EVENT, "security_arm_state_event"],
            [EventType.TIMER_EVENT, "timer_event"],
            [EventType.WEATHER_EVENT, "weather_event"],
        ],
    )
    def test_for_type(self, event_type, attribute):
        value = {"k": "v"}
        args = {"event_type": event_type, attribute: value}
        assert (Event(**args).for_type(event_type)) is value


class TestInstallRequest:
    def test_config_convenience_methods(self):
        data = load_file(FIXTURE_DIR / "live" / "request" / "INSTALL.1.json")
        request = CONVERTER.from_json(data, InstallRequest)
        assert request.as_str("retrieve-weather-enabled") == "true"
        assert request.as_bool("retrieve-weather-enabled") is True
        assert request.as_str("retrieve-weather-frequency") == "15"
        assert request.as_int("retrieve-weather-frequency") == 15
        assert request.as_float("retrieve-weather-frequency") == 15.0
        assert request.as_devices("humidity-devices") == [
            DeviceValue(device_id="3ac74985-XXXX-XXXX-XXXX-ea9623be6a7b", component_id="main"),
            DeviceValue(device_id="0ff440ec-XXXX-XXXX-XXXX-a39e189b8cc9", component_id="main"),
        ]


class TestUpdateRequest:
    def test_config_convenience_methods(self):
        data = load_file(FIXTURE_DIR / "samples" / "request" / "UPDATE.json")
        request = CONVERTER.from_json(data, UpdateRequest)
        assert request.as_str("minutes") == "5"
        assert request.as_bool("minutes") is True
        assert request.as_int("minutes") == 5
        assert request.as_float("minutes") == 5.0
        assert request.as_devices("contactSensor") == [
            DeviceValue(device_id="e457978e-5e37-43e6-979d-18112e12c961", component_id="main"),
        ]


class TestEventRequest:
    def test_for_type_device(self):
        data = load_file(FIXTURE_DIR / "samples" / "request" / "EVENT-DEVICE.json")
        request = CONVERTER.from_json(data, EventRequest)
        for event_type in [event_type for event_type in EventType if event_type != EventType.DEVICE_EVENT]:
            assert request.event_data.for_type(event_type) == []
        assert request.event_data.for_type(EventType.DEVICE_EVENT) == [DEVICE_EVENT]

    def test_for_type_timer(self):
        data = load_file(FIXTURE_DIR / "samples" / "request" / "EVENT-TIMER.json")
        request = CONVERTER.from_json(data, EventRequest)
        for event_type in [event_type for event_type in EventType if event_type != EventType.TIMER_EVENT]:
            assert request.event_data.for_type(event_type) == []
        assert request.event_data.for_type(EventType.TIMER_EVENT) == [TIMER_EVENT]

    def test_filter_device(self):
        data = load_file(FIXTURE_DIR / "samples" / "request" / "EVENT-DEVICE.json")
        request = CONVERTER.from_json(data, EventRequest)
        for event_type in [event_type for event_type in EventType if event_type != EventType.DEVICE_EVENT]:
            assert request.event_data.filter(event_type) == []
            assert request.event_data.filter(event_type, predicate=lambda _: False) == []
            assert request.event_data.filter(event_type, predicate=lambda _: True) == []
        assert request.event_data.filter(EventType.DEVICE_EVENT) == [DEVICE_EVENT]
        assert request.event_data.filter(EventType.DEVICE_EVENT, predicate=lambda _: False) == []
        assert request.event_data.filter(EventType.DEVICE_EVENT, predicate=lambda _: True) == [DEVICE_EVENT]
        assert request.event_data.filter(
            EventType.DEVICE_EVENT, predicate=lambda x: x["deviceId"] == "6f5ea629-4c05-4a90-a244-cc129b0a80c3"
        ) == [DEVICE_EVENT]

    def test_filter_timer(self):
        data = load_file(FIXTURE_DIR / "samples" / "request" / "EVENT-TIMER.json")
        request = CONVERTER.from_json(data, EventRequest)
        for event_type in [event_type for event_type in EventType if event_type != EventType.TIMER_EVENT]:
            assert request.event_data.filter(event_type) == []
            assert request.event_data.filter(event_type, predicate=lambda _: False) == []
            assert request.event_data.filter(event_type, predicate=lambda _: True) == []
        assert request.event_data.filter(EventType.TIMER_EVENT) == [TIMER_EVENT]
        assert request.event_data.filter(EventType.TIMER_EVENT, predicate=lambda _: False) == []
        assert request.event_data.filter(EventType.TIMER_EVENT, predicate=lambda _: True) == [TIMER_EVENT]
        assert request.event_data.filter(EventType.TIMER_EVENT, predicate=lambda x: x["name"] == "lights_off_timeout") == [
            TIMER_EVENT
        ]



# vim: set ft=python ts=4 sw=4 expandtab:

"""
Unit test utilities.
"""

from pathlib import Path


def load_file(path: Path) -> str:
    """Load text of a single file as a string."""
    return path.read_text(encoding="utf-8")


def load_dir(path: Path) -> dict[str, str]:
    """Load text of all files in a directory into a dict."""
    return {f.name: f.read_text(encoding="utf-8") for f in path.iterdir() if f.is_file()}
